# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

import string, time
from datetime import datetime

from nive.utils.utils import ConvertToDateTime
from nive.definitions import ConfigurationError, OperationalError




class DatabaseManager(object):
    """
    Base class for specific database servers
    
    SQL and administration calls are bases on MySql
    """

    modifyColumns = True

    def __init__(self):
        self.db = None
        self.dbConn = None
        self.useUtf8 = True

    # Database Options ---------------------------------------------------------------------

    def Connect(self, databaseName = "", inIP = "", inUser = "", inPW = ""):
        return self.db != None


    def SetDB(self, inDB):
        self.dbConn = inDB
        self.db = self.dbConn.cursor()
        return self.db != None


    def Close(self):
        if self.db:
            self.db.close()
        if self.dbConn:
            self.dbConn.close()
        self.db = None


    def IsDB(self):
        return self.db is not None


    # Options ----------------------------------------------------------------

    def ConvertDate(self, date):
        if not isinstance(date, datetime):
            date = ConvertToDateTime(date)
        if not date:
            return ""
        return date.strftime("%Y-%m-%d %H:%M:%S")


    def UpdateStructure(self, tableName, structure, modify = None, createIdentity = True):
        """
        modify = None to skip changing existing columns, list of column ids to modify if changed
        """
        # check table exists
        if not self.IsTable(tableName):
            if not self.CreateTable(tableName, columns=structure, createIdentity=createIdentity):
                return False
            # delay until table is created
            time.sleep(0.3)
            aCnt = 1
            while not self.IsTable(tableName):
                time.sleep(0.5)
                aCnt += 1
                if aCnt == 10:
                    raise OperationalError("timeout create table")
            return True

        columns = self.GetColumns(tableName, structure)

        for col in structure:
            name = col["id"]
            options = self.ConvertConfToColumnOptions(col)

            # skip id column
            if createIdentity and name == "id":
                continue

            if not self.IsColumn(tableName, name):
                if not self.CreateColumn(tableName, name, options):
                    return False
                continue

            if not modify:
                continue

            if not name in modify:
                continue

            # modify column settings
            if name in columns and columns[name].get("db"):
                if not self.ModifyColumn(tableName, name, "", options):
                    return False
            else:
                if not self.CreateColumn(tableName, name, options):
                    return False

        if createIdentity:
            if not self.IsColumn(tableName, "id"):
                if not self.CreateIdentityColumn(tableName, "id"):
                    return False
            
        return True


    def ConvertConfToColumnOptions(self, conf):
        """
        field representation:
        {"id": "type", "datatype": "list", "size"/"maxLen": 0, "default": ""}

        datatypes:
        list -> list || listn
        
        Example:

        string -> VARCHAR(size) NOT NULL DEFAULT default
        number -> INT NOT NULL DEFAULT default
        float -> FLOAT NOT NULL DEFAULT default
        bool -> TINYINT(4) NOT NULL DEFAULT default
        percent -> TINYINT(4) NOT NULL DEFAULT default
        text -> TEXT NOT NULL DEFAULT default
        htext -> TEXT NOT NULL DEFAULT default
        lines -> TEXT NOT NULL DEFAULT default
        xml -> TEXT NOT NULL DEFAULT default
        unit -> INT NOT NULL DEFAULT default
        unitlist -> VARCHAR(2048) NOT NULL DEFAULT default
        date -> DATE NULL DEFAULT default
        datetime -> DATETIME NULL DEFAULT default
        timestamp -> TIMESTAMP
        listt -> VARCHAR(30) NOT NULL DEFAULT default
        listn -> SMALLINT NOT NULL DEFAULT default
        multilist, checkbox -> VARCHAR(2048) NOT NULL DEFAULT default
        json -> TEXT
        [file] -> BLOB
        [bytesize] -> BIGINT(20) NOT NULL DEFAULT default
        url -> TEXT NOT NULL DEFAULT default
        """
        datatype = conf["datatype"]
        aStr = ""

        # convert datatype list
        
        return aStr


    # Database ------------------------------------------------------------

    def IsDatabase(self, databaseName):
        for aD in self.GetDatabases():
            if aD[0] == databaseName:
                return True
        return False

    def CreateDatabase(self, databaseName):
        if not self.IsDB():
            return False
        self.db.execute("create database " + databaseName)
        self.dbConn.commit()
        return True


    def UseDatabase(self, databaseName):
        if not self.IsDB():
            return False
        #self.db.execute("use " + databaseName)
        return True


    # Table --------------------------------------------------------------
    """
    inTableEntrys --> ColumnName ColumnDataTyp ColumnOptions
    ColumnOptions --> [NOT][NULL][DEFAULT = ""][PRIMARY KEY][UNIQUE]

    ENGINE = MyISAM
    CHARACTER SET utf8 COLLATE utf8_general_ci;

    ENGINE = InnoDB
    CHARACTER SET utf8 COLLATE utf8_general_ci;
    """

    def IsTable(self, tableName):
        for aD in self.GetTables():
            if aD[0].lower() == tableName.lower():
                return True
        return False


    def CreateTable(self, tableName, columns = None, createIdentity = True, primaryKeyName="id"):
        if not self.IsDB():
            return False
        if not tableName or tableName == "":
            return False
        assert(tableName != "user")
        aSql = ""
        if createIdentity:
            aSql = "CREATE TABLE %s(%s INT UNSIGNED AUTO_INCREMENT UNIQUE NOT NULL PRIMARY KEY" % (tableName, primaryKeyName)
            if columns:
                for c in columns:
                    if c.id == primaryKeyName:
                        continue
                    aSql += ","
                    aSql += c.id + " " + self.ConvertConfToColumnOptions(c)
            aSql += ")"
            aSql += " AUTO_INCREMENT = 1 "
        else:
            aSql = "CREATE TABLE %s" % (tableName)
            if not columns:
                raise ConfigurationError("No database fields defined.")
            aCnt = 0
            aSql += "("
            for c in columns:
                if aCnt:
                    aSql += ","
                aCnt = 1
                aSql += c.id + " " + self.ConvertConfToColumnOptions(c)
            aSql += ")"
        aSql += " ENGINE = %s" %(self.engine)
        if self.useUtf8:
            aSql += " CHARACTER SET utf8 COLLATE utf8_general_ci"

        self.db.execute(aSql)
        # delay until table is created
        time.sleep(0.3)
        aCnt = 1
        while not self.IsTable(tableName):
            time.sleep(0.5)
            aCnt += 1
            if aCnt == 30:
                return False
        return True

    def RenameTable(self, inTableName, inNewTableName):
        if not self.IsDB():
            return False
        self.db.execute("alter table %s rename as %s" % (inTableName, inNewTableName))
        return True


    def DeleteTable(self, inTableName):
        if not self.IsDB():
            return False
        self.db.execute("drop table %s" % (inTableName))
        return True

    # Columns --------------------------------------------------------------
    """
    inColumnData --> "ColumnName ColumnDatdatatype ColumnOptions"
        Column Options --> [Not] [NULL] [DEFAULT 'default']
    SetPrimaryKey&SetUnique --> can also get a ColumnList
        ColumnList --> ColumnName1,ColumName2,...
    """

    def IsColumn(self, tableName, columnName):
        columns = self.GetColumns(tableName)
        cn = columnName
        if cn in columns and columns[cn].get("db"):
            return True
        return False


    def CreateColumn(self, tableName, columnName, columnOptions):
        if not self.IsDB():
            return False
        #print("alter table %s add column %s %s" % (tableName, columnName, columnOptions))
        if columnOptions == "" or columnName == "" or tableName == "":
            return False
        if columnOptions == "identity":
            return self.CreateIdentityColumn(tableName, columnName)
        self.db.execute("alter table %s add column %s %s" % (tableName, columnName, columnOptions))
        return True


    def CreateIdentityColumn(self, tableName, columnName):
        if not self.IsDB():
            return False
        return self.CreateColumn(tableName, columnName, "INT UNSIGNED AUTO_INCREMENT UNIQUE NOT NULL PRIMARY KEY")


    def ModifyColumn(self, tableName, columnName, inNewColumnName, columnOptions):
        if not self.IsDB():
            return False
        #print("alter table %s change %s %s %s" % (tableName, columnName, inNewColumnName, columnOptions))
        if columnOptions == "" or columnName == "" or tableName == "":
            return False
        aN = inNewColumnName
        if aN == "":
            aN = columnName
        self.db.execute("alter table %s change %s %s %s" % (tableName, columnName, aN, columnOptions))
        return True


    # physical database structure ----------------------------------------------------------------

    def GetDatabases(self):
        if not self.IsDB():
            return ""
        self.db.execute("show databases")
        return self.db.fetchall()


    def GetTables(self):
        if not self.IsDB():
            return ""
        self.db.execute("show tables")
        return self.db.fetchall()


    def GetColumns(self, tableName, structure=None):
        """
        returns a dict of columns. each column is represented by and stored und column id:
        {db: {id, type, identity, default, null}, conf: FieldConf()}
        """
        if not self.IsDB():
            return []
        if not self.IsTable(tableName):
            return []
        self.db.execute("show columns from %s" % (tableName))
        table = {}
        for c in self.db.fetchall():
            table[c[0]] = {"db": {"id": c[0], "type": c[1], "identity": c[3]!="", "default": c[4], "null": ""}, 
                           "conf": None} 
        if structure:
            for conf in structure:
                if conf.id in table:
                    table[conf.id]["conf"] = conf
                else:
                    table[conf.id] = {"db": None, "conf": conf} 
        return table





# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

import threading
import logging
from time import time

# The following is used in *ConnectionRequest* classes to cache the current db connection
# for the time of a request. if pyramid is not available the thread local stack is used
# as fallback
try:
    from pyramid.threadlocal import get_current_request
except:
    def get_current_request():
        return None

from nive.definitions import OperationalError



class Connection(object):
    """
    Base database connection class. Parameter depend on database version.

    configuration (usage depends on database type):
    user = database user
    pass = password user
    host = host server
    port = port server
    dbName = database name
    """
    
    placeholder = "%s"

    def __init__(self, config = None, connectNow = True):
        self.configuration=config
        self.db = None
        self._vtime = time()
        if connectNow:
            self.connect()

    def __del__(self):
        self.close()

    # dbapi like functions --------------------------------------------------------------

    def cursor(self):
        db = self._get()
        if db is None:
            raise OperationalError("Database is closed")
        return db.cursor()
    
    def begin(self):
        """ Starts a new transaction, if supported """
        return

    def rollback(self):
        """ Calls rollback on the current transaction, if supported """
        db = self._get()
        if db:
            db.rollback()

    def commit(self):
        """ Calls commit on the current transaction, if supported """
        db = self._get()
        db.commit()

    def connect(self):
        """ Close and connect to server """
        # "use a subclassed connection"

    def close(self):
        """ Close database connection """
        db = self._get(connect=False)
        if db is not None:
            # rollback uncommited values by default
            db.rollback()
            db.close()
            self._set(None)
    
    def ping(self):
        """ ping database server """
        db = self._get()
        return db.ping()


    def IsConnected(self):
        """ Check if database is connected """
        try:
            db = self._get()
            return db.cursor()!=None
        except:
            return False

    def VerifyConnection(self):
        """ 
        reconnects if not connected. If revalidate is larger than 0 IsConnected() will only be called 
        after `revalidate` time
        """
        db = self._get()
        if not db:
            return self.connect()
        conf = self.configuration
        if not conf.verifyConnection:
            return True
        if conf.revalidate > 0:
            if self._getvtime()+conf.revalidate > time():
                return True
        if not self.IsConnected():
            return self.connect()
        self._setvtime()
        return True
    
    @property
    def dbapi(self):
        """ returns the dbapi database connection """
        return self._get() or self.connect()

    def PrivateConnection(self):
        """ """
        return None

    
    def FmtParam(self, param):
        """ 
        Format a parameter for sql queries like literal for db. This function is not
        secure for any values. 
        """
        if isinstance(param, (int, float)):
            return str(param)
        d = str(param)
        if d.find('"')!=-1:
            d = d.replace('"','\\"')
        return '"%s"'%d


    def GetDBManager(self):
        """ returns the database manager obj """
        raise TypeError("please use a subclassed connection")


    def _get(self, connect=True):
        # get stored database connection
        return self.db
        
    def _set(self, dbconn):
        # locally store database connection
        self.db = dbconn

    def _getvtime(self):
        # get stored database connection
        return self._vtime
        
    def _setvtime(self):
        # locally store database connection
        self._vtime = time()


class ConnectionThreadLocal(Connection):
    """
    Caches database connections as thread local values.
    """

    def __init__(self, config = None, connectNow = True):
        self.local = threading.local()
        Connection.__init__(self, config, False)
        
    def _get(self, connect=True):
        # get stored database connection
        if not hasattr(self.local, "db"):
            return None
        return self.local.db
        
    def _set(self, dbconn):
        # locally store database connection
        self.local.db = dbconn
        
    def _getvtime(self):
        if not hasattr(self.local, "_vtime"):
            self._setvtime()
        return self.local._vtime
        
    def _setvtime(self):
        self.local._vtime = time()


class ConnectionCache(object):
    """
    Supports caching of multiple different database connection
    """
    def __init__(self):
        self.connections = {}

    def __setitem__(self, key, conn):
        self.connections[key] = conn
    
    def __getitem__(self, key):
        return self.connections[key]
    
    def close(self):
        for key, conn in self.connections.items():
            # rollback uncommited values by default
            conn.rollback()
            conn.close()

class ConnectionRequest(Connection):
    """
    Caches database connections as webob request values. Uses thread local stack as fallback (e.g testing).
    """

    def __init__(self, config = None, connectNow = True):
        self.local = threading.local()
        Connection.__init__(self, config, False)

    def VerifyConnection(self):
        """ 
        Connection is verified when added to the request. Failures during requests are not to be 
        expected.
        """
        req = get_current_request()
        if req:
            # connect moved to _get()
            return True
        # fallback if no request available
        db = self._get()
        if not db:
            return self.connect()
        conf = self.configuration
        if not conf.verifyConnection:
            return True
        if conf.revalidate > 0:
            if self._getvtime()+conf.revalidate > time():
                return True
        if not self.IsConnected():
            return self.connect()
        return True
    
    
    def _get(self, connect=True):
        # get stored database connection
        req = get_current_request()
        if not req:
            # use thread local stack as fallback
            if not hasattr(self.local, "db"):
                return None
            return self.local.db
        try:
            db = req.__nive_db__[self.configuration.dbName]
            if not connect:
                return db
            return db or self.connect()
        except (AttributeError, TypeError, KeyError) as e:
            if not connect:
                return None
            return self.connect()
        
    def _set(self, dbconn):
        # store database connection
        req = get_current_request()
        if not req:
            # use thread local stack as fallback
            self.local.db = dbconn
            return
        if not hasattr(req, "__nive_db__"):
            req.__nive_db__ = ConnectionCache()
        req.__nive_db__[self.configuration.dbName] = dbconn

        def CloseDatabase(request):
            try:
                request.__nive_db__.close()
                request.__nive_db__ = None
            except AttributeError:
                pass
        req.add_finished_callback(CloseDatabase)
        
    def _getvtime(self):
        if not hasattr(self.local, "_vtime"):
            self._setvtime()
        return self.local._vtime
        
    def _setvtime(self):
        self.local._vtime = time()




# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = """
Owner Group extension module
----------------------------
Security extension to handle local group assginments for users.
Can be used for Roots and Objects or any other python class supporting events
and id attribute (number). Uses idhash for root objects.
"""

from nive.definitions import ModuleConf, Conf
from nive.definitions import implementer, ILocalGroups


@implementer(ILocalGroups)
class OwnerGroup(object):
    """
    """
    _owner = "group:owner"

    def GetLocalGroups(self, username, user=None):
        """
        Group assignments use the user name.
        returns a list of all local user groups, including parent settings
        """
        if username==self.meta.pool_createdby:
            return [self._owner]
        return []


    def AllLocalGroups(self):
        """
        returns a list of all local user group settings as list including 
        [username, group, id]. This function does not include parent level
        settings.
        """
        return [self._owner]


    def AddOwner(self, user, **kw):
        """
        Set the current user as group:owner in pool_createdby
        """
        if user is None:
            return
        self.meta["pool_createdby"] = str(user)

        
    def AddLocalGroup(self, username, group):
        """
        Add a local group assignment for username.
        """
        pass

        
    def RemoveLocalGroups(self, username, group=None):
        """
        Remove a local group assignment. If group is None all local groups
        will be removed.
        """
        pass


    def RemoveGroups(self, **kw):
        """
        Remove all group assignments before deleting the object. 
        """
        pass



def SetupOwnerGroup(app, pyramidConfig):
    # get all roots and add extension
    extension = "nive.extensions.owner.OwnerGroup"
    def add(confs):
        for c in confs:
            e = c.extensions
            if e and extension in e:
                continue
            if e is None:
                e = []
            if isinstance(e, tuple):
                e = list(e)
            e.append(extension)
            c.unlock()
            c.extensions = tuple(e)
            c.lock()
    
    add(app.configurationQuery.GetAllRootConfs())
    add(app.configurationQuery.GetAllObjectConfs())
    
    
configuration = ModuleConf(
    id = "localGroups",
    name = "Owner Group assignment for objects and roots",
    context = "nive.extensions.owner.OwnerGoup",
    events = (Conf(event="startRegistration", callback=SetupOwnerGroup),),
)



#-*- coding: utf-8 -*-

import unittest

from datetime import datetime
from datetime import time as datetime_time

from nive import File
from nive.definitions import ContainmentError

from nive.tests import db_app
from nive.tests import __local


class objTest_db:
    
    def setUp(self):
        self._loadApp()

    def tearDown(self):
        self._closeApp(True, True)


    def test_obj1(self):
        a=self.app
        r=db_app.root(a)
        user = db_app.User("test")
        o = db_app.createObj1(r)
        id = o.id
        o.Close()
        o = r.obj(id)
        self.assertTrue(o)
        self.assertTrue(o.app)
        self.assertTrue(o.root)
        self.assertTrue(o.db)
        self.assertTrue(o.IsRoot()==False)
        self.assertTrue(o.GetParents())
        self.assertTrue(o.GetParentIDs())
        self.assertTrue(o.GetParentTitles())
        self.assertTrue(o.GetParentPaths())
        self.assertTrue(o.relpath)
        o.Close()
        r.Delete(id, user=user)


    def test_obj_timezone(self):
        return
        a=self.app
        r=db_app.root(a)
        user = db_app.User("test")
        o = db_app.createObj1(r)
        id = o.id
        self.assertTrue(o.meta.pool_create)
        self.assertTrue(o.meta.pool_create.hour==datetime.utcnow().hour)
        o.Update(db_app.data1_2, user)
        self.assertTrue(o.meta.pool_change)
        self.assertTrue(o.meta.pool_change.hour==datetime.utcnow().hour)
        o.Close()

        r.Delete(id, user=user)


    def test_objectacl(self):
        #print "Testing object update and commit"
        a=self.app
        ccc = a.db.GetCountEntries()
        # create
        user = db_app.User("test")

        r=db_app.root(a)
        self.assertTrue(r.__acl__[0][2]=='view')

        oo = db_app.createObj1(r)
        self.assertTrue(oo.__acl__[0][2]=='view')

        r.Delete(oo, user=user)


    def test_objectedit(self):
        #print "Testing object update and commit"
        a=self.app
        ccc = a.db.GetCountEntries()
        r=db_app.root(a)
        # create
        user = db_app.User("test")

        oo = db_app.createObj1(r)
        o1 = db_app.createObj1(oo)
        self.assertTrue(oo.GetTypeID()=="type1")
        self.assertTrue(oo.GetTypeName()=="Type 1 container")
        self.assertTrue(oo.GetFieldConf("ftext").name=="ftext")
        self.assertTrue(oo.GetTitle()=="")
        self.assertTrue(oo.GetPath()==str(oo.id))
        self.assertTrue(str(o1.GetFld("ftime"))==str(datetime_time(16,55)), str(o1.GetFld("ftime")))

        id = o1.GetID()
        o1.Update(db_app.data1_2, user)
        self.assertTrue(o1.GetID()==id)
        self.assertTrue(o1.GetFld("ftext")==db_app.data1_2["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data1_2["pool_filename"])
        self.assertTrue(str(o1.GetFld("ftime"))==str(datetime_time(23,8,13,500)), str(o1.GetFld("ftime")))
        o1.Close()
        del o1

        o1 = oo.obj(id)
        self.assertTrue(o1.GetFld("ftext")==db_app.data1_2["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data1_2["pool_filename"])
        o1.Close()
        oo.Delete(id, user=user)

        o1 = db_app.createObj1(oo)
        id = o1.GetID()
        data, meta, files = o1.SplitData(db_app.data1_2)
        o1.meta.update(meta)
        o1.data.update(data)
        o1.files.update(files)
        o1.Commit(user)
        self.assertTrue(o1.GetID()==id)
        self.assertTrue(o1.GetFld("ftext")==db_app.data1_2["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data1_2["pool_filename"])
        del o1
        o1 = oo.obj(id)
        self.assertTrue(o1.GetFld("ftext")!=db_app.data1_1["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")!=db_app.data1_1["pool_filename"])
        
        o1.UpdateInternal(db_app.data1_1)
        o1.CommitInternal(user)
        self.assertTrue(o1.GetFld("ftext")==db_app.data1_1["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data1_1["pool_filename"])

        self.assertRaises(ContainmentError, r.Delete, id, user)
        r.Delete(oo.GetID(), user=user)
        self.assertEqual(ccc, a.db.GetCountEntries())
        a.Close()


    def test_objectfiles(self):
        #print "Testing object files update and commit"
        a=self.app
        r=db_app.root(a)
        ccc = a.db.GetCountEntries()
        # create
        user = db_app.User("test")

        # testing StoreFile()
        o1 = db_app.createObj2(r)
        o2 = db_app.createObj2(r)
        o1.Update(db_app.data2_1, user)
        o2.Update(db_app.data2_2, user)
        o1.StoreFile("file1", File(**db_app.file2_1), user=user)
        o1.StoreFile("file2", File(**db_app.file2_2), user=user)
        o2.StoreFile("file2", File(**db_app.file2_2), user=user)
        self.assertTrue(o1.GetFld("ftext")==db_app.data2_1["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_1["pool_filename"], o1.GetFld("pool_filename"))
        self.assertTrue(o1.GetFile("file1").filename==db_app.file2_1["filename"])
        self.assertTrue(o1.GetFile("file2").filename==db_app.file2_2["filename"])
        self.assertTrue(o2.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o2.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o2.GetFile("file2").filename==db_app.file2_2["filename"])

        id1 = o1.id
        id2 = o2.id
        del o1
        del o2
        o1 = r.obj(id1)
        o2 = r.obj(id2)
        self.assertTrue(o1.GetFld("ftext")==db_app.data2_1["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_1["pool_filename"])
        self.assertTrue(o1.GetFile("file1").filename==db_app.file2_1["filename"])
        self.assertTrue(o1.GetFile("file2").filename==db_app.file2_2["filename"])
        self.assertTrue(o2.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o2.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o2.GetFile("file2").filename==db_app.file2_2["filename"])
        self.assertTrue(o2.GetFileByName(db_app.file2_2["filename"]))

        #delete files
        self.assertTrue(o1.DeleteFile("file1", user=user))
        self.assertTrue(o1.DeleteFile("file2", user=user))
        self.assertFalse(o1.GetFile("file1"))
        self.assertFalse(o1.GetFile("file2"))
        self.assertFalse(o1.GetFileByName(db_app.file2_2["filename"]))

        r.Delete(o1.GetID(), user=user)
        r.Delete(o2.GetID(), user=user)

        # testing files wrapper
        o1 = db_app.createObj2(r)
        o2 = db_app.createObj2(r)
        d = db_app.data2_1.copy()
        d["file1"] = File(**db_app.file2_1)
        d["file2"] = File(**db_app.file2_2)
        o1.Update(d, user)
        d = db_app.data2_2.copy()
        d["file2"] = File(**db_app.file2_2)
        o2.Update(d, user)
        self.assertTrue(o1.GetFld("ftext")==db_app.data2_1["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_1["pool_filename"])
        self.assertTrue(o1.GetFile("file1").filename==db_app.file2_1["filename"])
        self.assertTrue(o1.GetFile("file2").filename==db_app.file2_2["filename"])
        self.assertTrue(o2.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o2.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o2.GetFile("file2").filename==db_app.file2_2["filename"])

        id1 = o1.id
        id2 = o2.id
        del o1
        del o2
        o1 = r.obj(id1)
        o2 = r.obj(id2)
        self.assertTrue(o1.GetFld("ftext")==db_app.data2_1["ftext"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_1["pool_filename"])
        self.assertTrue(o1.GetFile("file1").filename==db_app.file2_1["filename"])
        self.assertTrue(o1.GetFile("file2").filename==db_app.file2_2["filename"])
        self.assertTrue(o2.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o2.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o2.GetFile("file2").filename==db_app.file2_2["filename"])

        r.Delete(id1, user=user)
        r.Delete(id2, user=user)
        self.assertEqual(ccc, a.db.GetCountEntries())
        a.Close()


    def test_commitundo(self):
        #print "Testing object commit and undo for data and files"
        a=self.app
        ccc = a.db.GetCountEntries()
        r=db_app.root(a)
        # create
        user = db_app.User("test")

        # testing commit
        o1 = db_app.createObj2(r)
        d = db_app.data2_2.copy()
        d["file1"] = File(**db_app.file2_1)
        d["file2"] = File(**db_app.file2_2)
        data, meta, files = o1.SplitData(d)
        o1.data.update(data)
        o1.meta.update(meta)
        o1.files.update(files)
        self.assertTrue(o1.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o1.files["file1"]["filename"]==db_app.file2_1["filename"])
        self.assertTrue(o1.files["file2"]["filename"]==db_app.file2_2["filename"])
        o1.Commit(user)
        self.assertTrue(o1.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o1.GetFile("file1").filename==db_app.file2_1["filename"])
        self.assertTrue(o1.GetFile("file2").filename==db_app.file2_2["filename"])
        id = o1.id
        del o1
        o1 = r.obj(id)
        self.assertTrue(o1.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o1.GetFile("file1").filename==db_app.file2_1["filename"])
        self.assertTrue(o1.GetFile("file2").filename==db_app.file2_2["filename"])
        o1.Close()
        r.Delete(o1.GetID(), user=user)

        # testing undo
        o1 = db_app.createObj2(r)
        o1.StoreFile("file1", File(**db_app.file2_1), user=user)
        d = db_app.data2_2.copy()
        d["file1"] = File(**db_app.file2_2)
        data, meta, files = o1.SplitData(d)
        o1.data.update(data)
        o1.meta.update(meta)
        o1.files.update(files)
        self.assertTrue(o1.GetFld("fstr")==db_app.data2_2["fstr"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_2["pool_filename"])
        self.assertTrue(o1.files.get("file1").filename==db_app.file2_2["filename"])
        o1.Undo()
        self.assertTrue(o1.GetFld("fstr")==db_app.data2_1["fstr"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_1["pool_filename"])
        self.assertTrue(o1.files.get("file1").filename==db_app.file2_1["filename"])
        id = o1.id
        del o1
        o1 = r.obj(id)
        self.assertTrue(o1.GetFld("fstr")==db_app.data2_1["fstr"])
        self.assertTrue(o1.GetFld("pool_filename")==db_app.data2_1["pool_filename"])
        self.assertTrue(o1.GetFile("file1").filename==db_app.file2_1["filename"])
        o1.Close()
        r.Delete(id, user=user)
        self.assertEqual(ccc, a.db.GetCountEntries())
        a.Close()

    
class objTest_db_sqlite(objTest_db, __local.SqliteTestCase):
    """
    see tests.__local
    """

class objTest_db_mysql(objTest_db, __local.MySqlTestCase):
    """
    see tests.__local
    """
        
class objTest_db_pg(objTest_db, __local.PostgreSqlTestCase):
    """
    see tests.__local
    """


class groupsTest_db:
    
    def setUp(self):
        self._loadApp(["nive.extensions.localgroups"])

    def tearDown(self):
        self._closeApp(True)


    def test_objectgroups(self):
        a=self.app
        r=db_app.root(a)
        # create
        user = db_app.User("test")
        o = db_app.createObj1(r)
        id = o.id
        
        userid = "test"
        #self.assertEqual(o.GetLocalGroups(userid), ["group:owner"]) # TODO MySQL failure pycharm?

        r.RemoveLocalGroups(userid, None)
        o.RemoveLocalGroups(userid, None)
        #r.db.Commit()
        self.assertFalse(o.GetLocalGroups(userid))
        o.AddLocalGroup(userid, "group:local")
        self.assertEqual(o.GetLocalGroups(userid), ["group:local"])
        o.RemoveLocalGroups("nouser", "nogroup")
        self.assertEqual(o.GetLocalGroups(userid), ["group:local"])
        o.RemoveLocalGroups(userid, "nogroup")
        self.assertEqual(o.GetLocalGroups(userid), ["group:local"])
        o.RemoveLocalGroups("nouser", "group:local")
        self.assertEqual(o.GetLocalGroups(userid), ["group:local"])
        o.RemoveLocalGroups(userid, "group:local")
        self.assertFalse(o.GetLocalGroups(userid))
        o.AddLocalGroup(userid, "group:local")
        o.RemoveLocalGroups(userid, None)
        self.assertFalse(o.GetLocalGroups(userid))
        o.Commit(user=user)

        r.Delete(id, user=user)


class groupsTest_db_sqlite(groupsTest_db, __local.SqliteTestCase):
    """
    see tests.__local
    """

class groupsTest_db_mysql(groupsTest_db, __local.MySqlTestCase):
    """
    see tests.__local
    """
        
class groupsTest_db_pg(groupsTest_db, __local.PostgreSqlTestCase):
    """
    see tests.__local
    """
        
 
    
#TODO tests obj tool + workflow

class objToolTest_db:

    """
    def test_tools(self):
        GetTool(name)
    """


class objWfTest_db:

    """
    def test_wf(self):
        GetWorkflow()
        GetNewWfID()
        GetWfStateName(user)
        GetWfInfo(user)
        WfAction(action, transition = None, user = None)
        WfAllow(action, transition = None, user = None)
        SetWfp(processID, user, force=False)
        SetWfState(stateID, user)
        SetWfData(state, key, data, user)
        GetWfData(state = None, key = None)
        DeleteWfData(user, state = None, key = None)
        GetWfLog(lastEntryOnly=1)
        AddWfLog(action, transition, user, comment="")
    """






import time
import unittest
import datetime

from nive.definitions import ObjectConf, FieldConf, ViewModuleConf
from nive.utils.path import DvPath
from nive.helper import *

# -----------------------------------------------------------------

class text(object):
    pass

class text1(object):
    pass
class text2(object):
    pass
class text3(object):
    pass

testconf = ObjectConf(
    id = "text",
    name = "Text",
    dbparam = "texts",
    context = "nive.tests.test_helper.text",
    selectTag = 1,
    description = ""
)

testconf.data = (
    FieldConf(id="textblock", datatype="htext", size=10000, default="", name="Text", fulltext=True, description="")
)

testconf.forms = {
    "create": { "fields":  ["textblock"],
                "actions": ["create"]},
    "edit":   { "fields":  ["textblock"], 
                "actions": ["save"]}
}
configuration = testconf


class EncoderTest(unittest.TestCase):

    def test_jsonencoder(self):
        self.assertTrue(JsonDataEncoder().encode({})=="{}")
        self.assertTrue(JsonDataEncoder().encode({"a":"a","b":1}))
        self.assertTrue(JsonDataEncoder().encode({"a":datetime.now()}))
        self.assertTrue(JsonDataEncoder().encode({"a":File()}))

    def test_confencoder(self):
        self.assertTrue(ConfEncoder().encode(Conf()))
        self.assertTrue(ConfEncoder().encode(Conf(**{"a":"a","b":1})))
        self.assertTrue(ConfEncoder().encode(ObjectConf()))
        self.assertTrue(ConfEncoder().encode(ObjectConf(**{"id":"a","name":"1"})))

    # automatic class exports as ccc not supported yet
    #def test_confdecoder(self):
    #    c=ConfEncoder().encode(Conf())
    #    self.assert_(ConfDecoder().decode(c)!=None)
    #    c=ConfEncoder().encode(Conf(**{"a":"a","b":1}))
    #    self.assert_(ConfDecoder().decode(c).a=="a",c)
    #    c=ConfEncoder().encode(ObjectConf())
    #    self.assert_(ConfDecoder().decode(c),c)
    #    c=ConfEncoder().encode(ObjectConf(**{"id":"a","name":"1"}))
    #    self.assert_(ConfDecoder().decode(c).id=="a",c)


class ConfigurationTest(unittest.TestCase):

    def test_resolve1(self, **kw):
        self.assertTrue(ResolveName("nive.tests.test_helper.text", base=None))
        self.assertTrue(ResolveName(".test_helper.text", base="nive.tests"))
        self.assertTrue(ResolveName("nive.tools.dbStructureUpdater", base=None))
        self.assertTrue(ResolveName(".dbStructureUpdater.dbStructureUpdater", base="nive.tools"))
        
    def test_resolve2(self, **kw):
        i,c = ResolveConfiguration(testconf, base=None)
        self.assertTrue(c)
        i,c = ResolveConfiguration("nive.tests.test_helper.testconf", base=None)
        self.assertTrue(c)
        i,c = ResolveConfiguration(".test_helper.testconf", base="nive.tests")
        self.assertTrue(c)
        i,c = ResolveConfiguration("nive.tests.test_helper", base=None)
        self.assertTrue(c)

    def test_load1(self, **kw):
        p=DvPath(__file__)
        p.SetNameExtension("app.json")
        i,c = ResolveConfiguration(str(p))
        self.assertTrue(c)

    def test_load2(self, **kw):
        p=DvPath(__file__)
        p.SetNameExtension("db.json")
        c = LoadConfiguration(str(p))
        self.assertTrue(c)
        
        
    def test_classref(self):
        #self.assert_(GetClassRef(testconf.context, reloadClass=False, raiseError=False)==text)
        self.assertTrue(GetClassRef(text, reloadClass=False, raiseError=False)==text)
        self.assertTrue(GetClassRef("nive.tests.test_helper.textxxxxxx", reloadClass=False, raiseError=False)==None)
        self.assertRaises(ImportError, GetClassRef, "nive.tests.test_helper.textxxxxxx", reloadClass=False, raiseError=True)
        

    def test_factory(self):
        #self.assert_(ClassFactory(testconf, reloadClass=False, raiseError=False)==text)
        c=testconf.copy()
        c.context=text
        self.assertTrue(ClassFactory(c, reloadClass=False, raiseError=False)==text)
        c=testconf.copy()
        c.context="nive.tests.test_helper.textxxxxxx"
        self.assertTrue(ClassFactory(c, reloadClass=False, raiseError=False)==None)
        c=testconf.copy()
        c.context="nive.tests.test_helper.textxxxxxx"
        self.assertRaises(ImportError, ClassFactory, c, reloadClass=False, raiseError=True)

        c=testconf.copy()
        c.extensions=(text1,text2,text3)
        self.assertTrue(len(ClassFactory(c, reloadClass=False, raiseError=False).mro())==6)
        c=testconf.copy()
        c.extensions=("text66666",text2,"text5555")
        self.assertTrue(len(ClassFactory(c, reloadClass=False, raiseError=False).mro())==4)
        c=testconf.copy()
        c.extensions=("nive.tests.test_helper.text1","nive.tests.test_helper.text2","nive.tests.test_helper.text3")
        self.assertTrue(len(ClassFactory(c, reloadClass=False, raiseError=False).mro())==6)


    def test_replace(self):
        alist = [Conf(id="1", value=123),Conf(id="2", value=456),Conf(id="3", value=789)]
        repl = Conf(id="2", value=999)
        new = ReplaceInListByID(alist, repl)
        self.assertTrue(new!=alist)
        self.assertTrue(new[1].value==999)


    def test_update(self):
        alist = [Conf(id="1", value=123),Conf(id="2", value=456),Conf(id="3", value=789)]
        repl = Conf(id="2", value=999)
        new = UpdateInListByID(alist, repl)
        self.assertTrue(new==alist)
        self.assertTrue(new[1].value==999)


    def test_decorator(self):
        class someclass(object):
            pass
        conf = ViewModuleConf(id="test")
        newcls = DecorateViewClassWithViewModuleConf(conf, someclass)
        self.assertTrue(newcls.__configuration__)
        self.assertTrue(newcls.__configuration__().id=="test")

        newcls = DecorateViewClassWithViewModuleConf(conf, "nive.tests.test_helper.text")
        self.assertTrue(newcls.__configuration__)
        self.assertTrue(newcls.__configuration__().id=="test")


from nive.tests import db_app
from nive.tests import __local

class ListItemTest_db:

    def setUp(self):
        self._loadApp()

    def tearDown(self):
        self._closeApp(True)


    def test_listitems(self):
        val = LoadListItems(self.app.configurationQuery.GetFld("pool_type"), app=self.app, obj=None, pool_type=None, force=True)
        self.assertTrue(len(val)==2, val) # one type is hidden
        self.assertTrue(LoadListItems(FieldConf(id="test",datatype="list",settings={"codelist":"groups"}), app=self.app))
        self.assertTrue(LoadListItems(FieldConf(id="test",datatype="list",settings={"codelist":"languages"}), app=self.app))
        self.assertTrue(LoadListItems(FieldConf(id="test",datatype="list",settings={"codelist":"countries"}), app=self.app))
        self.assertTrue(LoadListItems(FieldConf(id="test",datatype="list",settings={"codelist":"types"}), app=self.app))
        # TODO fill database
        LoadListItems(FieldConf(id="test", datatype="list", settings={"codelist": "users"}), app=self.app)
        LoadListItems(FieldConf(id="test",datatype="list",settings={"codelist":"meta"}), app=self.app)
        LoadListItems(FieldConf(id="test",datatype="list",settings={"codelist":"type:type1"}), app=self.app)



class ListItemTest_db_sqlite(ListItemTest_db, __local.SqliteTestCase):
    """
    see tests.__local
    """

class ListItemTest_db_mysql(ListItemTest_db, __local.MySqlTestCase):
    """
    see tests.__local
    """

class ListItemTest_db_pg(ListItemTest_db, __local.PostgreSqlTestCase):
    """
    see tests.__local
    """



# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

from nive.i18n import _
from nive.utils.utils import ConvertListToStr, ConvertToNumberList

from nive.definitions import IContainer, IRoot, ISort

"""
Cut, copy and paste functionality for objects 

conf for objects:
    - self.disableCopy: either unset or True/False

"""



class CopyView:
    """
    View functions for cut, copy and paste
    """
    CopyInfoKey = "CCP__"

    def cut(self):
        """
        """
        self.ResetFlashMessages()
        ids = self.GetFormValue("ids")
        if not ids:
            ids = [self.context.id]
        cp = self.SetCopyInfo("cut", ids, self.context)
        url = self.GetFormValue("url")
        if not url:
            url = self.PageUrl(self.context)
        msgs = _("OK. Cut.")
        ok = True
        return self.Redirect(url, [msgs], refresh=True)


    def copy(self):
        """
        """
        self.ResetFlashMessages()
        ids = self.GetFormValue("ids")
        if not ids:
            ids = [self.context.id]
        cp = self.SetCopyInfo("copy", ids, self.context)
        url = self.GetFormValue("url")
        if not url:
            url = self.PageUrl(self.context)
        msgs = _("OK. Copied.")
        return self.Redirect(url, [msgs], refresh=True)


    def paste(self):
        """
        """
        self.ResetFlashMessages()
        deleteClipboard=1
        url = self.GetFormValue("url")
        if not url:
            url = self.PageUrl(self.context)
        action, ids = self.GetCopyInfo()
        if not action or not ids:
            msgs = []
            return self.Redirect(url, msgs, refresh=True)

        pepos = self.GetFormValue("pepos",0)
        result = False
        msgs = [_("Method unknown")]
        if action == "cut":
            result, msgs = self.Move(ids, pepos, user=self.User())
            if result and deleteClipboard:
                cp = self.DeleteCopyInfo()
        elif action == "copy":
            result, msgs = self.Paste(ids, pepos, user=self.User())
        return self.Redirect(url, msgs, refresh=result)
    
    
    def SetCopyInfo(self, action, ids, context):
        """
        store in session or cookie
        """
        if not ids:
            return ""
        if isinstance(ids, str):
            ids=ConvertToNumberList(ids)
        cp = ConvertListToStr([action]+ids).replace(" ","")
        self.request.session[self.CopyInfoKey] = cp
        return cp


    def GetCopyInfo(self):
        """
        get from session or cookie
        """
        cp = self.request.session.get(self.CopyInfoKey,"")
        if isinstance(cp, str):
            cp = cp.split(",")
        if not cp or len(cp)<2:
            return "", []
        return cp[0], cp[1:]

    
    def ClipboardEmpty(self):
        """
        check if clipboard is empty
        """
        cp = self.request.session.get(self.CopyInfoKey,"")
        return cp==""
    

    def DeleteCopyInfo(self):    
        """
        reset copy info
        """
        self.request.session[self.CopyInfoKey] = ""


    def CanCopy(self, context=None):
        """
        """
        context = context or self.context
        if IRoot.providedBy(context):
            return False
        return not hasattr(context, "disableCopy") or not context.disableCopy

    def CanPaste(self, context=None):
        """
        """
        context = context or self.context
        if IContainer.providedBy(context):
            return False
        return not hasattr(context, "disableCopy") or not context.disableCopy


    def Paste(self, ids, pos, user, context=None):
        """
        Paste the copied object with id to this object
        """
        context = context or self.context
        root = context.root
        new = []
        msgs = []
        result = True
        for id in ids:
            id = int(id)
            if context.id == id:
                continue
            obj = root.LookupObj(id, preload="skip")
            if not obj:
                msgs.append(_("Object not found"))
                result = False
                continue
            newobj = context.Duplicate(obj, user)
            if not newobj:
                raise TypeError("Duplicate failed")
            if ISort.providedBy(context):
                context.InsertAfter(newobj.id, pos, user=user)
            new.append(newobj)
        if not context.app.configuration.autocommit:
            for o in new:
                o.Commit(user)
        if result:
            msgs.append(_("OK. Copied and pasted."))
        return result, msgs

    def Move(self, ids, pos, user, context=None):
        """
        Move the object with id to this object

        Events

        - beforeAdd(data=obj.meta, type=type)
        - afterDelete(id=obj.id)
        - moved()
        - afterAdd(obj=obj)
        """
        context = context or self.context
        root = context.root
        oldParent = None

        moved = []
        msgs = []
        result = True
        for id in ids:
            id = int(id)
            if context.id == id:
                continue
            obj = root.LookupObj(id, preload="skip")
            if not obj:
                msgs.append(_("Object not found"))
                result = False
                continue

            type = obj.GetTypeID()
            # allow subobject
            if not context.IsTypeAllowed(type, user):
                raise TypeError("Object cannot be added here")

            context.Signal("beforeAdd", data=obj.meta, type=type)
            if not oldParent or oldParent.id != obj.parent.id:
                oldParent = obj.parent
            obj.__parent__ = context
            obj.meta["pool_unitref"] = context.id
            oldParent.Signal("afterDelete", id=obj.id)
            obj.Signal("moved")
            # obj.Close()
            moved.append(obj)

        for o in moved:
            o.Commit(user)
            if ISort.providedBy(context):
                self.InsertAfter(o.id, pos, user=user)
            context.Signal("afterAdd", obj=o)
        if result:
            msgs.append(_("OK. Cut and pasted."))
        return result, msgs






import os
import sys

from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))

try:
    README = open(os.path.join(here, 'readme.md')).read()
    CHANGES = open(os.path.join(here, 'changes.txt')).read()
except:
    README = ''
    CHANGES = ''

requires = [
    'Chameleon==3.6.2', # form error in 3.8.1
    'zope.interface',
    'pyramid>=1.10.1',
    'pyramid_chameleon',
    'iso8601',
    'pytz',
    'translationstring',
    'bs4',    # alternative - 'html2text'
    # 'pillow' optional. required for extensions.images.
]

setupkw = dict(
      name='nive',
      version='1.3.15',
      description='Nive 3 base package',
      long_description=README + '\n\n' + CHANGES,
      long_description_content_type="text/markdown",
      classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.7"
      ],
      author='Arndt Droullier, Nive GmbH',
      author_email='info@nive.co',
      url='https://niveapps.com/',
      keywords='cms framework pyramid',
      license='GPL 3',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      tests_require=requires,
      test_suite="nive"
)

setup(**setupkw)



# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = """
This file provides global *search* functionality, database lookup and sql query 
wrappers. It is usually attached to root objects.

Search parameter handling
--------------------------
Most functions support a similar parameter handling for generating sql queries.  

===========  ===========================================================================
Parameters
===========  ===========================================================================
fields       list of database fields (field ids) to include in result.
             prefix '-': special fields or aggregate functions can be inserted by adding 
                         a '-' in front of the field. -COUNT(*) or -MAX(field1)
             prefix '+': special fields skipped in SQL query e.g. '+preview'
parameter    dictionary with fieldname:value entries used for search conditions
operators    dictionary with fieldname:operator entries used for search conditions
             default: strings=*LIKE*, all others='='
             possible values: ``=, LIKE, NOT IN, IN, >, <, <=, >=, !=, BETWEEN ``
sort         result sort field or list if multiple
ascending    sort ascending or decending
start        start position in result
max          maximum number of result records
===========  ===========================================================================

Parameter for operator BETWEEN has to be formatted: e.g. '2006/10/10' AND '2006/10/11'

================  =====================================================================
Keyword options   (Not supported by all functions)
================  =====================================================================
logicalOperator   link between conditions. default: *AND*. possible: ``AND, OR, NOT``
condition         custom sql condition statement appended to WHERE clause 
groupby           used as sql GROUP BY statement
join              adds a custom custom sql join statement
jointype          left, inner, right. default = inner (used by ``SearchType()``)
mapJoinFld        fieldname to map title by left/right join (used by ``SearchType()``)
addID             Add fld.id in query without showing in result list. used for custom 
                  columns e.g. -(select ...)
skipRender        render result flds as html element. 
                  default: ``pool_type, pool_wfa, pool_wfp`` 
                  to skip all: *True*, or a list of fields  ("pool_wfa","pool_type")
skipCount         enable or disable second query to get the number of all records.
relation          Resolve a relation and load the related entry as object. For example
                  you load the user object by settings `relation=pool_createdby`. The
                  result set will include the user object instead of the user name.
================  =====================================================================

Adding custom join statements

    "join" = "inner join pool_meta as __parent on __parent.id = meta__.pool_unitref"

    fields = ["-__parent.pool_filename as profile_filename"]


=============  ========================================================================
Search result  (supported by Search*() functions)
=============  ========================================================================
criteria       dictionary containing parameters used in search
count          number of records contained in result
total          total number of records matching the query
items          result list. each record as dictionary. Each field is rendered for 
               display depending on *skipRender* keyword. This means list entries are
               replaced by their readable names, dates are rendered readable. 
start          record start number in result 
max            maximum number of records 
next           start number of next record set
nextend        end number of next record set
prev           start number of previous record set
prevend        end number of previous record set
sql            the sql statement used
=============  ========================================================================

"""

import time

from nive.utils.utils import ConvertToNumberList
from nive.views import FieldRenderer
from nive.definitions import IFieldConf
from nive.definitions import FieldConf
from nive.definitions import ConfigurationError, ConnectionError


class Search(object):
    """ Provides search functionality  """

    def __init__(self, root):
        self.root = root

    @property
    def app(self):
        return self.root.app

    @property
    def db(self):
        return self.root.db


    # Simple search functions ----------------------------------------------------------------------------------------------
    
    def Select(self, pool_type=None, parameter=None, fields=None, operators=None, sort=None, ascending=1, start=0, max=0, **kw):
        """
        Fast and simple sql query. 
        
        If *pool_type* is not set the query will apply to the meta layer only. In this case you can
        only include pool_meta.fields in *fields*. To use a single table pass *dataTable* as keyword.
        
        If *pool_type* is set you can use meta and data *fields*. The query is restricted to a single 
        type.
        
        Supported keywords: ``groupby, logicalOperator, condition, dontAddType, dataTable, db``
        
        The following example selects all children of the current object ::
            
            fields = ["id", "title", "pool_type"]
            parameter["pool_unitref"] = self.id
            records = self.root.search.Select(parameter=parameter, fields=fields)
        
        returns records as list
        """
        parameter = parameter or {}
        operators = operators or {}
        fields = fields or ["id"]
        db = kw.get("db") or self.db
        if pool_type is None:
            dataTable=kw.get("dataTable") or "pool_meta"
            sql, values = db.FmtSQLSelect(fields, 
                                          parameter, 
                                          dataTable=dataTable, 
                                          singleTable=1, 
                                          operators=operators, 
                                          sort=sort, 
                                          ascending=ascending, 
                                          start=start, 
                                          max=max, 
                                          groupby=kw.get("groupby"), 
                                          logicalOperator=kw.get("logicalOperator"), 
                                          condition=kw.get("condition"))
        else:
            if "pool_type" not in parameter and not kw.get("dontAddType"):
                parameter["pool_type"] = pool_type
            if "pool_type" not in operators:
                operators["pool_type"] = "="
            typeInf = self.app.configurationQuery.GetObjectConf(pool_type)
            if not typeInf:
                raise ConfigurationError(pool_type + " type not found")
            sql, values = db.FmtSQLSelect(fields, 
                                          parameter, 
                                          dataTable=typeInf["dbparam"], 
                                          operators=operators, 
                                          sort=sort, 
                                          ascending=ascending, 
                                          start=start, 
                                          max=max, 
                                          groupby=kw.get("groupby"), 
                                          logicalOperator=kw.get("logicalOperator"), 
                                          condition=kw.get("condition"))
        recs = db.Query(sql, values)
        return recs


    def SelectDict(self, pool_type=None, parameter=None, fields=None, operators=None, sort=None, ascending=1, start=0, max=0, **kw):
        """
        Fast and simple sql query. 
        
        If *pool_type* is not set the query will apply to the meta layer only. In this case you can
        only include pool_meta.fields in *fields*. 
        
        If *pool_type* is set you can use meta and data *fields*. The query is restricted to a single 
        type.
        
        Supported keywords: ``groupby, logicalOperator, condition``
        
        Records are returned as dictionaries
        
        The following example selects all children *not* of type *image* of the current object ::
            
            fields = ["id", "title", "pool_type"]
            parameter = {"pool_unitref": self.id}
            operators = {"pool_type": "!="}
            records = self.root.search.SelectDict("image",
                                               parameter=parameter, 
                                               fields=fields, 
                                               operators=operators)
        
        returns records as dict list
        """
        fields = fields or ["id"]
        recs = self.Select(pool_type=pool_type,
                           parameter=parameter, 
                           fields=fields, 
                           operators=operators, 
                           sort=sort, 
                           ascending=ascending, 
                           start=start, 
                           max=max, 
                           **kw)
        if len(recs)==0:
            return recs
        # convert
        if len(fields) > len(recs[0]):
            raise TypeError("Too many fields")
        for i in range(len(fields)):
            name = fields[i]
            if name.find(" as ")!=-1:
                name = name.split(" as ")[1]
                fields[i] = name
        return [dict(list(zip(fields, r))) for r in recs]


    # Extended search functions ----------------------------------------------------------------------------------------------
    
    def Search(self, parameter, fields=None, operators=None, **kw):
        """
        Extended meta layer search function. Supports all keyword options and search result. 
        
        Example ::
        
            root.()Search({"title":"test"}, 
                          fields=["id", "pool_type", "title"], 
                          start=0, max=50, 
                          operators={"title":"="})
        
        returns search result (See above)
        """
        t = time.time()
        fields = fields or []
        operators = operators or {}
        start, max, kw = self._SearchKWs(kw)
        # lookup field definitions
        fields, fldList, groupcol = self._PrepareFields(fields)
        removeID = self._HandleGroupByQueries(fields, fldList, groupcol, kw)

        db = self.db
        if not db:
            raise ConnectionError("No database connection")
        
        sql, values = db.FmtSQLSelect(fldList, 
                                      parameter=parameter, 
                                      operators=operators, 
                                      start=start,
                                      max=max,
                                      **kw)
        records = db.Query(sql, values)

        # convert records
        fldList = self._RenameFieldAlias(fldList)
        converter = self._PrepareRenderer(kw, fldList)
        items, cnt = self._ConvertRecords(records, converter, fields, fldList, kw)

        # total records
        total = len(items) + start
        if (total==max or start>0) and kw.get("skipCount") != 1:
            if "sort" in kw:
                del kw["sort"]
            if not kw.get("groupby"):
                sql2, values = db.FmtSQLSelect(["-count(*)"], 
                                               parameter=parameter, 
                                               operators=operators, 
                                               start=None, 
                                               max=None,
                                               **kw)
                val = db.Query(sql2, values)
                total = val[0][0] if val else 0
            else:
                sql2, values = db.FmtSQLSelect(["-count(DISTINCT %s)" % (kw.get("groupby"))], 
                                               parameter=parameter, 
                                               operators=operators, 
                                               start=None, 
                                               max=None, 
                                               **kw)
                val = db.Query(sql2, values)
                total = len(val) if val else 0

        items = self._HandleRelations(kw.get("relations"), items, kw)
        result = self._PrepareResult(items, parameter, cnt, total, start, max, t, sql)
        return result


    def SearchType(self, pool_type, parameter=None, fields=None, operators=None, **kw):
        """
        Extended meta and data layer search function. Supports all keyword options and search result. 
        
        Example ::
        
            root.()SearchType("image", 
                              parameter={"title":"test"}, 
                              fields=["id", "pool_type", "title"], 
                              start=0, max=50, 
                              operators={"title":"="})
        
        returns search result (See above)
        """
        t = time.time()
        fields = fields or []
        operators = operators or {}
        start, max, kw = self._SearchKWs(kw)
        # lookup field definitions
        fields, fldList, groupcol = self._PrepareFields(fields, pool_type)
        removeID = self._HandleGroupByQueries(fields, fldList, groupcol, kw)
        self._HandleTypeJoins(pool_type, parameter, operators, kw)

        typeInf = self.app.configurationQuery.GetObjectConf(pool_type)
        if not typeInf:
            raise ConfigurationError("Type not found (%s)" % (pool_type))

        db = self.db
        if db is None:
            raise ConnectionError("No database connection")

        sql, values = db.FmtSQLSelect(fldList, 
                                      parameter=parameter,  
                                      operators=operators,
                                      start=start, 
                                      max=max, 
                                      dataTable=typeInf["dbparam"],
                                      **kw)
        records = db.Query(sql, values)
            
        # prepare field renderer and names
        fldList = self._RenameFieldAlias(fldList)
        converter = self._PrepareRenderer(kw, fldList)
        items, cnt = self._ConvertRecords(records, converter, fields, fldList, kw)

        # total records
        total = len(items) + start
        if (total==max or start>0) and kw.get("skipCount") != 1:
            if "sort" in kw:
                del kw["sort"]
            if not kw.get("groupby"):
                cntflds = ["-count(*) as cnt"]
            else:
                cntflds = ["-count(DISTINCT %s) as cnt" % (kw.get("groupby"))]
            sql2, values = db.FmtSQLSelect(cntflds, 
                                           parameter=parameter,  
                                           operators=operators,
                                           dataTable=typeInf["dbparam"],
                                           **kw)
            val = db.Query(sql2, values)
            if not kw.get("groupby"):
                total = val[0][0] if val else 0
            else:
                total = len(val) if val else 0

        items = self._HandleRelations(kw.get("relations"), items, kw)
        result = self._PrepareResult(items, parameter, cnt, total, start, max, t, sql)
        return result


    def SearchData(self, pool_type, parameter, fields=None, operators=None, **kw):
        """
        Extended data layer search function. Supports all keyword options and search result. 
        
        Example ::
        
            root.()SearchData("image", 
                              parameter={"text": "new"}, 
                              fields=["id", "text", "title"], 
                              start=0, max=50, 
                              operators={"text":"LIKE"})
        
        returns search result (See above)
        """
        t = time.time()
        fields = fields or []
        operators = operators or {}
        start, max, kw = self._SearchKWs(kw)
        # lookup field definitions
        fields, fldList, groupcol = self._PrepareFields(fields, pool_type)
        removeID = self._HandleGroupByQueries(fields, fldList, groupcol, kw)

        typeInf = self.app.configurationQuery.GetObjectConf(pool_type)
        if not typeInf:
            raise ConfigurationError("Type not found (%s)" % (pool_type))

        db = self.db
        if not db:
            raise ConnectionError("No database connection")

        sql, values = db.FmtSQLSelect(fldList, 
                                      parameter=parameter,  
                                      operators=operators,
                                      start=start, 
                                      max=max, 
                                      dataTable=typeInf["dbparam"],
                                      singleTable=1,
                                      **kw)
        records = db.Query(sql, values)
            
        # prepare field renderer and names
        fldList = self._RenameFieldAlias(fldList)
        converter = self._PrepareRenderer(kw, fldList)
        items, cnt = self._ConvertRecords(records, converter, fields, fldList, kw)

        # total records
        total = len(items) + start
        if (total==max or start>0) and kw.get("skipCount") != 1:
            if "sort" in kw:
                del kw["sort"]
            if not kw.get("groupby"):
                cntflds = ["-count(*) as cnt"]
            else:
                cntflds = ["-count(DISTINCT %s) as cnt" % (kw.get("groupby"))]
            sql2, values = db.FmtSQLSelect(cntflds, 
                                           parameter=parameter,  
                                           operators=operators,
                                           start=start, 
                                           max=max, 
                                           dataTable=typeInf["dbparam"],
                                           singleTable=1,
                                           **kw)
            val = db.Query(sql2, values)
            if not kw.get("groupby"):
                total = val[0][0] if val else 0
            else:
                total = len(val) if val else 0
                
        items = self._HandleRelations(kw.get("relations"), items, kw)
        result = self._PrepareResult(items, parameter, cnt, total, start, max, t, sql)
        return result


    def SearchFulltext(self, phrase, parameter=None, fields=None, operators=None, **kw):
        """
        Fulltext search function. Searches all text fields marked for fulltext search. Uses *searchPhrase* 
        as parameter for text search. Supports all keyword options and search result. 
        
        Example ::
        
            root.SearchFulltext("new", parameter={},
                              fields=["id", "title"], 
                              start=0, max=50, 
                              operators={"text":"LIKE"})
        
        returns search result (See above)
        """
        t = time.time()

        fields = fields or ("id","title","pool_type","-pool_fulltext.text as fulltext")
        operators = operators or {}
        parameter = parameter or {}
        if phrase==None:
            phrase = ""
        searchFor = phrase
        start, max, kw = self._SearchKWs(kw)
        # lookup field definitions
        fields, fldList, groupcol = self._PrepareFields(fields)
        removeID = self._HandleGroupByQueries(fields, fldList, groupcol, kw)

        if phrase.find("*") == -1:
            phrase = "%%%s%%" % phrase
        else:
            phrase = phrase.replace("*", "%")

        db = self.db
        if not db:
            raise ConnectionError("No database connection")

        sql, values = db.GetFulltextSQL(phrase, 
                                        fldList, 
                                        parameter=parameter,  
                                        operators=operators, 
                                        start=start, 
                                        max=max,
                                        **kw)
        records = db.Query(sql, values)
            
        # prepare field renderer and names
        fldList = self._RenameFieldAlias(fldList)
        converter = self._PrepareRenderer(kw, fldList)
        items, cnt = self._ConvertRecords(records, converter, fields, fldList, kw)
        
        # total records
        total = len(items) + start
        if (total==max or start>0) and kw.get("skipCount") != 1:
            if "sort" in kw:
                del kw["sort"]
            if not kw.get("groupby"):
                cntflds = ["-count(*) as cnt"]
            else:
                cntflds = ["-count(DISTINCT %s) as cnt" % (kw.get("groupby"))]
            sql2, values = db.FmtSQLSelect(cntflds, 
                                           parameter=parameter,  
                                           operators=operators,
                                           start=start, 
                                           max=max, 
                                           skipRang=1, 
                                           **kw)
            val = db.Query(sql2, values)
            if not kw.get("groupby"):
                total = val[0][0] if val else 0
            else:
                total = len(val) if val else 0

        items = self._HandleRelations(kw.get("relations"), items, kw)
        result = self._PrepareResult(items, parameter, cnt, total, start, max, t, sql)
        result["phrase"] = searchFor
        return result


    def SearchFulltextType(self, pool_type, phrase, parameter=None, fields=None, operators=None, **kw):
        """
        Fulltext search function. Searches all text fields marked for fulltext search of the given type. Uses *searchPhrase* 
        as parameter for text search. Supports all keyword options and search result. 
        
        Example ::
        
            root.()SearchFulltextType("text",
                              parameter={"searchPhrase": "new"}, 
                              fields=["id", "text", "title"], 
                              start=0, max=50, 
                              operators={"text":"LIKE"})
        
        returns search result (See above)
        """
        t = time.time()

        fields = fields or ("id","title","-pool_fulltext.text as fulltext")
        operators = operators or {}
        parameter = parameter or {}
        if phrase is None:
            phrase = ""
        searchFor = phrase
        start, max, kw = self._SearchKWs(kw)
        # lookup field definitions
        fields, fldList, groupcol = self._PrepareFields(fields, pool_type)
        removeID = self._HandleGroupByQueries(fields, fldList, groupcol, kw)
        self._HandleTypeJoins(pool_type, parameter, operators, kw)

        # fulltext wildcard
        if phrase.find("*") == -1:
            phrase = "%%%s%%" % phrase
        else:
            phrase = phrase.replace("*", "%")

        typeInf = self.app.configurationQuery.GetObjectConf(pool_type)
        if not typeInf:
            raise ConfigurationError("Type not found (%s)" % (pool_type))

        db = self.db
        if not db:
            raise ConnectionError("No database connection")

        sql, values = db.GetFulltextSQL(phrase, 
                                        fldList, 
                                        parameter=parameter, 
                                        operators=operators,
                                        start=start, 
                                        max=max,
                                        dataTable=typeInf["dbparam"],
                                        **kw)
        records = db.Query(sql, values)
            
        # prepare field renderer and names
        fldList = self._RenameFieldAlias(fldList)
        converter = self._PrepareRenderer(kw, fldList)
        items, cnt = self._ConvertRecords(records, converter, fields, fldList, kw)

        # total records
        total = len(items) + start
        if (total==max or start>0) and kw.get("skipCount") != 1:
            if "sort" in kw:
                del kw["sort"]
            if not kw.get("groupby"):
                cntflds = ["-count(*) as cnt"]
            else:
                cntflds = ["-count(DISTINCT %s) as cnt" % (kw.get("groupby"))]
            sql2, values = db.FmtSQLSelect(cntflds, 
                                           parameter=parameter,  
                                           operators=operators,
                                           start=start, 
                                           max=max, 
                                           dataTable=typeInf["dbparam"],
                                           skipRang=1, 
                                           **kw)
            val = db.Query(sql2, values)
            if not kw.get("groupby"):
                total = val[0][0] if val else 0
            else:
                total = len(val) if val else 0

        items = self._HandleRelations(kw.get("relations"), items, kw)
        result = self._PrepareResult(items, parameter, cnt, total, start, max, t, sql)
        result["phrase"] = searchFor
        return result


    def SearchFilename(self, filename, parameter, fields=None, operators=None, **kw):
        """
        Filename search function. Searches all physical file filenames (not url path names). Supports all 
        keyword options and search result. 
        
        Includes matchinng files as "result_files" in each record.
        
        Example ::
        
            root.()SearchFulltextType("text",
                              parameter={"searchPhrase": "new"}, 
                              fields=["id", "text", "title"], 
                              start=0, max=50, 
                              operators={"text":"LIKE"})
        
        returns search result (See above)
        """
        fields = fields or ()
        operators = operators or {}

        db = self.db
        if not db:
            raise ConnectionError("No database connection")
        files = db.SearchFilename(filename)

        ids = []
        filesMapped = {}
        for f in files:
            if not f["id"] in ids:
                ids.append(f["id"])
            if str(f["id"]) not in filesMapped:
                filesMapped[str(f["id"])] = []
            filesMapped[str(f["id"])].append(f)
        if len(ids)==0:
            ids.append(0)
        
        parameter["id"] = ids
        operators["id"] = "IN"
        result = self.Search(parameter=parameter, fields=fields, operators=operators, **kw)

        # merge result and files
        for rec in result["items"]:
            if str(rec["id"]) not in filesMapped:
                continue
            rec["result_files"] = filesMapped[rec["id"]]
            # copy first file in list
            file = filesMapped[str(rec["id"])][0]
            rec["+size"] = file["size"]
            rec["+extension"] = file["extension"]
            rec["+filename"] = file["filename"]
            rec["+fileid"] = file["fileid"]
            rec["+key"] = file["filekey"]

        # update result
        del result["criteria"]["id"]
        result["criteria"]["filename"] = filename
        return result


    def _SearchKWs(self, kws):
        default = {"ascending":1, "start":0, "max":100}
        # parse incoming kws
        
        try:       start = int(kws["start"])
        except:    start = default["start"]
        if "start" in kws:
            del kws["start"]

        try:       max = int(kws["max"])
        except:    max = default["max"]
        if "max" in kws:
            del kws["max"]

        try:       kws["ascending"] = int(kws["ascending"])
        except:    kws["ascending"] = default["ascending"]
        
        """
        groupby=kw.get("groupby"), 
        logicalOperator=kw.get("logicalOperator"), 
        condition=kw.get("condition"), 
        join=kw.get("join"))
        """
        return start, max, kws
         
    def _PrepareFields(self, fields, pool_type=None):
        # lookup field definitions
        fields = self._GetFieldDefinitions(fields, pool_type)
        fldList = []
        groupcol = 0
        for f in fields:
            if f["id"][0] == "-":
                groupcol += 1
            fldList.append(f["id"])
        return fields, fldList, groupcol
    
    def _GetFieldDefinitions(self, fields, pool_type = None):
        f = []
        for fld in fields:
            if IFieldConf.providedBy(fld):
                f.append(fld)
                continue
            if not isinstance(fld, str):
                continue
            if fld in ("__preview__",):
                continue
            # skip custom flds
            if fld[0] == "+":
                continue
            # Aggregate functions, custom flds
            if fld[0] == "-":
                f.append(FieldConf(**{"id": fld, "name": fld, "datatype": "string"}))
                continue

            fl = self.app.configurationQuery.GetFld(fld, pool_type)
            if fl:
                f.append(fl)
        return f

    def _HandleGroupByQueries(self, fields, fldList, groupcol, kws):
        # add id. required for group by  queries
        removeID = False
        if (not "id" in fldList and groupcol == 0 and kws.get("groupby") == None) or kws.get("addID")==1:
            fldList.append("id")
            fields.append(self.app.configurationQuery.GetFld("id"))
            removeID = True
        return removeID
    
    def _HandleTypeJoins(self, pool_type, parameter, operators, kws):
        # set join type
        default_join = 0
        if "jointype" not in kws or kws.get("jointype")=="inner":
            default_join = 1
            if not kws.get("skiptype"):
                parameter["pool_type"] = pool_type
        #operators
        if not operators:
            operators = {}
        if "pool_type" not in operators:
            operators["pool_type"] = "="
        if not default_join:
            operators["jointype"] = kws.get("jointype")
        return default_join

    
    def _PrepareRenderer(self, kws, fldList):
        # prepare field renderer
        skipRender = kws.get("skipRender", False)
        if skipRender == True:
            skipRender = fldList
        elif not skipRender:
            skipRender = ("pool_type", "pool_wfa", "pool_wfp")
        converter = FieldRenderer(self, skip=skipRender)
        return converter
    
    def _RenameFieldAlias(self, fldList):
        # parse alias field names used in sql query
        p = 0
        for f in fldList:
            if f[0] == "-" and f.find(" as ") != -1:
                a = f.split(" as ")[-1]
                a = a.replace(" ", "")
                a = a.replace(")", "")
                fldList[p] = a
            p += 1
        return fldList

    def _ConvertRecords(self, records, converter, fields, fldList, kws):
        # convert result
        db = self.db
        items = []
        for rec in records:
            rec2 = []
            for p in range(len(fields)):
                value = db.structure._de(rec[p], fields[p]["datatype"], fields[p])
                rec2.append(converter.Render(fields[p], value, False, **kws))
            items.append(dict(list(zip(fldList, rec2))))
        return items, len(items)
    
    def _HandleRelations(self, relations, items, kws):
        if not relations:
            return items
        if isinstance(relations, str):
            relations = (relations,)
        cache = {}
        for r in relations:
            # special cases: user lookup pool_createdby, pool_changedby
            if r in ("pool_createdby", "pool_changedby"):
                GetUser = self.app.portal.userdb.root.GetUser
                for i in items:
                    ref = "user:"+i[r]
                    if ref in cache:
                        user = cache[ref]
                    else:
                        user = GetUser(i[r])
                        cache[ref] = user
                    i[r] = user
            # todo load obj by id
        return items

    def _PrepareResult(self, items, parameter, cnt, total, start, max, t, sql):
        # prepare result dictionary and paging information
        result = {}
        result["items"] = items
        result["criteria"] = parameter
        result["count"] = cnt
        result["total"] = total
        result["start"] = start
        result["max"] = max
        result["time"] = time.time() - t
        result["sql"] = sql
        
        next = start + max
        if next >= total:
            next = 0
        result["next"] = next
        nextend = next + max
        if nextend > total:
            nextend = total
        result["nextend"] = nextend

        prev = start - max
        if prev < 1:
            prev = 0
        result["prev"] = prev
        prevend = prev + max
        result["prevend"] = prevend
        return result


    # Tree structure -----------------------------------------------------------

    def GetTree(self, flds=None, sort="id", base=0, parameter=""):
        """
        Select list of all folders from db.
        
        returns the subtree
        {'items': [{'id': 354956L, 'ref1': 354954L, 'ref2': 354952L, ..., 'ref10': None, 'items': [...]}]
        """
        if not flds:
            # lookup meta list default fields
            flds = self.app.configuration.listDefault
        db = self.db
        return db.GetTree(flds=flds, sort=sort, base=base, parameter=parameter)


    def TreeParentIDs(self, id):
        """
        returns the parent ids for the object with id as list
        """
        db = self.db
        return db.GetParentPath(id)


    def TreeParentTitles(self, id):
        """
        returns the parent titles for the object with id as list
        """
        db = self.db
        return db.GetParentTitles(id)


    # Codelists representation for entries -----------------------------------------------------------------------------------------

    def GetEntriesAsCodeList(self, pool_type, name_field, parameter=None, operators=None, sort=None):
        """
        Search the database for entries of type *pool_type* and return matches as codelist ::
        
            [{"name": name_field, "id": object.id}, ... ]
        
        If the name_field is stored as data field, insert *data.* at the beginning of name_field 
        (e.g. ``data.header``)
        
        returns list
        """
        operators = operators or {}
        parameter = parameter or {}
        if not sort:
            sort = name_field
        if "pool_type" not in parameter:
            parameter["pool_type"] = pool_type
        recs = self.SelectDict(pool_type=pool_type, parameter=parameter, 
                               fields=["id", "pool_unitref", name_field+" as name"], 
                               operators=operators, sort=sort)
        return recs


    def GetEntriesAsCodeList2(self, name_field, parameter=None, operators=None, sort=None):
        """
        Search the database and return matches as codelist ::
        
            [{"name": name_field, "id": object.id}, ... ]
        
        returns list
        """
        operators = operators or {}
        parameter = parameter or {}
        if not sort:
            sort = name_field
        recs = self.SelectDict(parameter=parameter, fields=["id", "pool_unitref", name_field+" as name"], 
                               operators=operators, sort=sort)
        return recs


    def GetGroupAsCodeList(self, pool_type, name_field, parameter=None, operators=None, sort=None):
        """
        Search the database for entries of type *pool_type* and return matches grouped by unique
        *name_field* values as codelist ::
        
            [{"name": name_field, "id": object.id}, ... ]
        
        If the name_field is stored as data field, insert *data.* at the beginning of name_field 
        (e.g. ``data.header``)
        
        returns list
        """
        operators = operators or {}
        parameter = parameter or {}
        if not sort:
            sort = name_field
        if "pool_type" not in parameter:
            parameter["pool_type"] = pool_type
        recs = self.SelectDict(pool_type=pool_type, 
                               parameter=parameter, 
                               fields=["id", "pool_unitref", name_field+" as name"], 
                               operators=operators, 
                               #groupby=name_field, 
                               sort=sort)
        return recs


    def GetGroupAsCodeList2(self, name_field, parameter=None, operators=None, sort=None):
        """
        Search the database and return matches grouped by unique *name_field* values as codelist ::
        
            [{"name": name_field, "id": object.id}, ... ]
        
        returns list
        """
        operators = operators or {}
        parameter = parameter or {}
        if not sort:
            sort = name_field
        recs = self.SelectDict(parameter=parameter, 
                               fields=["id", "pool_unitref", name_field+" as name"], 
                               operators=operators, 
                               #groupby=name_field, 
                               sort=sort)
        return recs


    # Name / ID lookup -----------------------------------------------------------

    def FilenameToID(self, filename, unitref=None, parameter=None, firstResultOnly=True, operators=None):
        """
        Convert url path filename (meta.pool_filename) to id. This function does not lookup
        physical files and their filenames.

        returns id
        """
        operators = operators or {}
        parameter = parameter or {}
        if unitref is not None:
            parameter["pool_unitref"] = unitref
        parameter["pool_filename"] = filename
        operators["pool_filename"] = "="
        # lookup meta list default fields
        flds = self.app.configuration.listDefault
        recs = self.Select(parameter=parameter, fields=flds, operators=operators)
        #print recs
        if firstResultOnly:
            if len(recs) == 0:
                return 0
            return recs[0][0]
        return recs


    def IDToFilename(self, id):
        """
        Convert id to url path filename (meta.pool_filename). This function does not lookup
        physical files and their filenames.
        
        returns string
        """
        parameter={"id": id}
        recs = self.Select(parameter=parameter, fields=["pool_filename"])
        if len(recs) == 0:
            return ""
        return recs[0][0]


    def ConvertDatarefToID(self, pool_type, dataref):
        """
        Search for object id based on dataref and pool_type.

        returns id
        """
        parameter={"pool_type": pool_type, "pool_dataref": dataref}
        recs = self.Select(parameter=parameter, fields=["id"])
        if len(recs) == 0:
            return 0
        return recs[0][0]


    def GetMaxID(self):
        """
        Lookup id of the last created object.

        returns id
        """
        recs = self.Select(fields=["-max(id)"])
        if len(recs) == 0:
            return 0
        return recs[0][0]


    # References --------------------------------------------------------------------

    def GetReferences(self, unitID, types=None, sort="id"):
        """
        Search for references in unit or unitlist fields of all objects.
        
        returns id list
        """
        if not types:
            types = self.app.configurationQuery.GetAllObjectConfs()
        else:
            l = []
            for t in types:
                l.append(self.app.configurationQuery.GetObjectConf(t))
            types = l
        references = []
        ids = [unitID]

        db = self.db

        # lookup meta list default fields
        #flds = self.app.configuration.listDefault
        #flds = list(flds)
        # search unit flds
        for t in types:
            for f in t["data"]:
                if f["datatype"] != "unit":
                    continue
                l = self.Select(pool_type=t["id"], parameter={f["id"]:unitID}, fields=["id"], sort=sort)
                for r in l:
                    if not r[0] in ids:
                        ids.append(r[0])
                        references.append(r)

        # search unitlist flds
        for t in types:
            for f in t["data"]:
                if f["datatype"] != "unitlist":
                    continue
                l = self.Select(pool_type=t["id"], parameter={f["id"]:str(unitID)}, fields=["id", f["id"]], sort=sort,
                                operators={f["id"]:"LIKE"})
                for r in l:
                    if not r[2] == t["id"]:
                        continue
                    unitRefs = ConvertToNumberList(r[3])
                    if not unitID in unitRefs:
                        continue
                    if not r[0] in ids:
                        ids.append(r[0])
                        references.append(r)

        return references





# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

import json
import base64

from nive.tool import Tool, ToolView
from nive.definitions import ToolConf, FieldConf, ViewConf
from nive.definitions import IApplication, MetaTbl, Structure, IContainer
from nive.helper import ConfEncoder
from nive.i18n import _

configuration = ToolConf(
    id = "exportJson",
    context = "nive.tools.exportJson.exportJson",
    name = _("Json data export"),
    description = _("This function exports all objects in json format. Optionally as flat list or tree structure. This is not a simple database table dump. Object events will be triggered before exporting data."),
    apply = (IApplication,),
    mimetype = "application/json",
    data = [
        FieldConf(id="tree", 
                  datatype="bool", 
                  default=1, 
                  listItems=[{"id":"true", "name":"Tree"},{"id":"false","name":"Flat list"}], 
                  name=_("Export as tree"),
                  description=_("Export objects as tree structure (contained objects are included as 'items')")),
        FieldConf(id="filedata", 
                  datatype="radio", 
                  default="none", 
                  listItems=[{"id":"none", "name":"Only file information (No file data)"},
                             {"id":"path", "name":"Only file information and local paths (No file data)"},
                             {"id":"data", "name":"Include all file data"},], 
                  name=_("File data"),
                  description=_("Include binary file data in json export (encoded as base64)")),
        FieldConf(id="tag", datatype="string", default="exportJson", hidden=1)
    ],
    views = [
        ViewConf(name="", view=ToolView, attr="form", permission="system", context="nive.tools.exportJson.exportJson"),
    ]
)

class exportJson(Tool):
    """
    """

    def _Run(self, **values):

        result = 1
        codepage="utf-8"
    
        self.InitStream()
        app = self.app
        datapool = app.db
        conf = app.dbConfiguration
        conn = datapool.connection
        tree = values.get("tree")
        filedata = values.get("filedata")
        self.filename = app.configuration.id + ".json"

        if not conn:
            self.stream.write(_("Database connection error (${name})\n", mapping={"name": app.dbConfiguration.context}))
            return self.stream, 0
        
        if not conn.IsConnected():
            self.stream.write(_("Database connection error (${name})\n", mapping={"name": app.dbConfiguration.context}))
            return self.stream, 0
        
        def mapfields(fields):
            return [f.id for f in fields]
        
        metaflds = mapfields(app.configurationQuery.GetAllMetaFlds(ignoreSystem=False))

        def exportObj(o):
            values = {"__items__": []}
            for field in metaflds:
                values[field] = o.meta.get(field)
            for field in o.configuration.data:
                if field["datatype"] == "file":
                    file = o.files.get(field.id)
                    if not file:
                        continue
                    values[field.id] = {"filename": file.filename, "size": file.size}
                    if filedata=="path":
                        values[field.id]["path"] = file.path
                    elif filedata=="data":
                        values[field.id]["data"] = base64.b64encode(file.read())
                else:
                    values[field.id] = o.data.get(field.id)
            
            if IContainer.providedBy(o):
                for child in o.GetObjs():
                    cv = exportObj(child)
                    values["__items__"].append(cv)
            return values
                
        root = app.root
        data = {"__items__": []}
        for child in root.GetObjs():
            cv = exportObj(child)
            data["__items__"].append(cv)
            
        if not tree:
            def getlist(values, datalist):
                for v in values["__items__"]:
                    datalist = getlist(v, datalist)
                del values["__items__"]
                datalist.append(values)
                return datalist
            data = getlist(data, [])
            
        self.stream.write(ConfEncoder().encode(data))
        
        return self.stream, 1





import time
import unittest

from nive.definitions import Conf
from nive.events import *



# -----------------------------------------------------------------

class testobj(Events):
    def __init__(self):
        self.called = 0
        self.InitEvents()
    
    def callme(self):
        self.Signal("callme", data=1)

    def event_testLocal(self, **kw):
        self.called = kw.get("data")
        

class EventTest(unittest.TestCase):

    def setUp(self):
        self.obj = testobj()


    def event_test(self, **kw):
        self.called = kw.get("data")

    def test_events1(self):
        self.obj.ListenEvent("test", "event_testLocal")
        self.obj.Signal("test", data=12345)
        self.assertTrue(self.obj.called==12345)
        self.obj.RemoveListener("test", "event_testLocal")
        self.obj.Signal("test", data=67890)
        self.assertTrue(self.obj.called==12345)

    def test_eventobj(self):
        self.called = 0
        self.obj.ListenEvent("callme", self.event_test)
        self.obj.callme()
        self.assertTrue(self.called==1)
        self.obj.RemoveListener("callme", self.event_test)
        self.called=0
        self.obj.callme()
        self.assertTrue(self.called==0)


    def test_eventcontext(self):
        
        def event_fnc_test(context=None, data=None):
            self.assertTrue(context==self.obj)
            self.called=1
        
        self.called = 0
        self.obj.ListenEvent("callme", event_fnc_test)
        self.obj.callme()
        self.assertTrue(self.called==1)
        self.obj.RemoveListener("callme", event_fnc_test)
        self.called=0
        self.obj.callme()
        self.assertTrue(self.called==0)


    def test_eventresult(self):
        
        def event_fnc_test1(context=None, data=None):
            self.assertTrue(context==self.obj)
            return data
        def event_fnc_test2(context=None, data=None):
            self.assertTrue(context==self.obj)
            return data
        
        self.called = 0
        self.obj.ListenEvent("callme", event_fnc_test1)
        result = self.obj.Signal("callme", data=1)
        self.assertTrue(len(result)==1)
        self.assertTrue(result[0][0]==1, result)

        self.obj.ListenEvent("callme", event_fnc_test2)
        result = self.obj.Signal("callme", data=1)
        self.assertTrue(len(result)==2, result)
        self.assertTrue(result[0][0]==1)
        self.assertTrue(result[1][0]==1)

        self.obj.RemoveListener("callme")


    def test_eventfromconf(self):

        def event_fnc_test1(context=None, data=None):
            self.assertTrue(context==self.obj)
            return data
        def event_fnc_test2(context=None, data=None):
            self.assertTrue(context==self.obj)
            return data

        events = [Conf(event="callme1",callback=event_fnc_test1),
                  Conf(event="callme2",callback=event_fnc_test2)]

        self.called = 0
        self.obj.SetupEventsFromConfiguration(events)

        result = self.obj.Signal("callme1", data=1)
        self.assertTrue(len(result)==1)
        self.assertTrue(result[0][0]==1, result)

        result = self.obj.Signal("callme2", data=2)
        self.assertTrue(len(result)==1, result)
        self.assertTrue(result[0][0]==2)

        self.obj.RemoveListener("callme1")
        self.obj.RemoveListener("callme2")



import unittest

class TestValidationFailure(unittest.TestCase):
    def _makeOne(self, field, cstruct, error):
        from nive.components.reform.exception import ValidationFailure
        return ValidationFailure(field, cstruct, error)

    def test_render(self):
        widget = DummyWidget()
        form = DummyForm(widget)
        cstruct = {}
        e = self._makeOne(form, cstruct, None)
        result = e.render()
        self.assertEqual(result, cstruct)

class DummyForm(object):
    def __init__(self, widget):
        self.widget = widget
    
class DummyWidget(object):
    def serialize(self, field, cstruct):
        return cstruct
    





import copy, time, io
import unittest
from pkg_resources import resource_filename

from nive.utils.dataPool2.files import File


class fileentrytest(object):
    test=1
    
    @property
    def root(self):
        return resource_filename('nive.utils.dataPool2', 'tests/')
    
    @property
    def pool(self):
        return self
    
    def abspath(self, file):
        root = resource_filename('nive.utils.dataPool2', 'tests/')
        return root+"test_files.py"


class FileTest(unittest.TestCase):

    def test_create(self):
        #   File( filekey="", 
        #         filename="", 
        #         file=None, 
        #         size=0, 
        #         path="", 
        #         extension="", 
        #         fileid=0, 
        #         uid="", 
        #         tempfile=False, 
        #         filedict=None, 
        #         fileentry=None)
        file = File(filekey="aaa",
                    filename="qqqq.png", 
                    file=b"0123456789"
                    )
        self.assertTrue(file.filekey=="aaa")
        self.assertTrue(file.filename=="qqqq.png")
        self.assertTrue(file.size==10)
        self.assertTrue(file.extension=="png")

        file = File(filekey="aaa",
                    filename="qqqq.png", 
                    file=b"0123456789",
                    fileentry=fileentrytest()
                    )
        self.assertTrue(file.filekey=="aaa")
        self.assertTrue(file.filename=="qqqq.png")
        self.assertTrue(file.size==10)
        self.assertTrue(file.extension=="png")

        file = File(filekey="aaa",
                    path="/tmp/qqqq.png" 
                    )
        self.assertTrue(file.filekey=="aaa")
        self.assertTrue(file.filename=="qqqq.png")
        self.assertTrue(file.extension=="png")
        
        file = File(filedict={"filekey":"aaa",
                              "filename":"qqqq.png", 
                              "file":b"0123456789",
                              "fileentry":fileentrytest()}
                    )
        self.assertTrue(file.filekey=="aaa")
        self.assertTrue(file.filename=="qqqq.png")
        self.assertTrue(file.size==10)
        self.assertTrue(file.extension=="png")


    def test_read(self):
        file = File("aaa", filename="import.zip", tempfile=True)
        self.assertTrue(file.isTempFile())
        self.assertTrue(file.filename=="import.zip")
        self.assertTrue(file.extension=="zip")
        self.assertRaises(IOError, file.read)

        file = File(filekey="aaa",
                    filename="qqqq.png", 
                    file=b"0123456789",
                    fileentry=fileentrytest()
                    )
        self.assertTrue(file.read()==b"0123456789")

        file = File(filekey="aaa",
                    filename="qqqq.png", 
                    file=b"0123456789",
                    fileentry=fileentrytest()
                    )
        self.assertTrue(file.read(5)==b"01234")
        self.assertTrue(file.read(5)==b"56789")
        self.assertTrue(file.read(5)==b"")
        self.assertTrue(file.tell()==10)
        file.seek(0)
        self.assertTrue(file.tell()==0)

        file = File(filekey="aaa",
                    file=None,
                    fileentry=fileentrytest()
                    )
        self.assertRaises(IOError, file.read)
        file.close()
        self.assertFalse(file.isTempFile())


    def test_path(self):
        root = resource_filename('nive.utils.dataPool2', 'tests/')
        file = File("aaa")
        file.fromPath(root+"test_db.py")
        self.assertTrue(file.filename=="test_db.py")
        self.assertTrue(file.extension=="py")
        self.assertFalse(file.abspath())
        file.close()

        file = File(filekey="aaa",
                    file=None
                    )
        file.fileentry=fileentrytest
        file.path = root+"test_db.py"
        self.assertTrue(file.abspath().startswith(root))
        
        file = File("aaa", filename="qqqq.png", size=10)
        file.fromPath(root+"test_db.py")
        self.assertTrue(file.filename=="qqqq.png")
        self.assertTrue(file.extension=="png")
        file.close()
        
        
    def test_dict(self):
        file = File(filekey="aaa",
                    filename="qqqq.png", 
                    file=b"0123456789",
                    fileentry=fileentrytest()
                    )
        self.assertTrue(file.get("filekey")=="aaa")
        self.assertTrue(file.get("filename")=="qqqq.png")
        self.assertTrue(file.get("size")==10)

        self.assertTrue(file["filekey"]=="aaa")
        self.assertTrue(file["filename"]=="qqqq.png")
        self.assertTrue(file["size"]==10)

        self.assertTrue(file.get("none",123)==123)
        
        file.update({"filekey":"bbb", "filename":"oooo.png"})
        self.assertTrue(file.get("filekey")=="bbb")
        self.assertTrue(file.get("filename")=="oooo.png")
        
        for k in file:
            self.assertTrue(k)




from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotFound

from nive.definitions import IObjectConf, IConf, Conf
from nive.definitions import IObject, IContainer, INonContainer, IRoot
from nive.definitions import ViewConf, ViewModuleConf

from nive.views import BaseView
from nive.components.reform.forms import ObjectForm, HTMLForm
from nive.security import Allow
from nive.helper import ResolveName
from nive.workflow import WorkflowNotAllowed

from .search import Search
from .parts import Parts
from .cutcopy import CopyView
from .definitions import AllTypeDataFlds, AllTypeMetaFlds

from nive.i18n import _, translator

# view module definition ------------------------------------------------------------------

#@nive_module
configuration = ViewModuleConf(
    id = "ifaceconf",
    name = "Pool management interface",
    containment = "nive.components.iface.definitions.IIFaceRoot",
    view = "nive.components.iface.view.IFaceView",
    static = (("s-iface","nive.components.iface:static/"),),
    templates = "nive.components.iface:templates/",
    template = "nive.components.iface:templates/index.pt",
    ifaceConf = None,
    permission = "view",
    acl = [
        (Allow, "group:editor", "iface_view"),
        (Allow, "group:editor", "iface_edit"), 
        (Allow, "group:editor", "iface_add"),
        (Allow, "group:editor", "iface_delete"),
        (Allow, "group:editor", "iface_wf"),
        (Allow, "group:editor", "iface_tools")
    ],
    assets = [
        ('bootstrap.min.css', 'nive.components.iface:static/mods/bootstrap-4.3.1-dist/css/bootstrap.css'),
        ('glyphicon.css', 'nive.components.iface:static/mods/glyphicons/css/glyphicon.css'),   # nive css
        #('adminview.css', 'nive.components.iface:static/adminview.css'),   # nive css
        ('iface.css', 'nive.components.iface:static/iface.css'),
        ('jquery.js', 'nive.components.iface:static/mods/jquery-3.3.1.min.js'),
        ('bootstrap.min.js', 'nive.components.iface:static/mods/bootstrap-4.3.1-dist/js/bootstrap.bundle.js'),
        ('iface.js', 'nive.components.iface:static/iface.js'),
    ]
)

# shortcuts
#_Object = "nive.components.iface.view.ObjectIFace"
#_Container = "nive.components.iface.view.ContainerIFace"
#_Root = "nive.components.iface.view.RootIFace"
t = configuration.templates

configuration.views = [
    # Root views --------------------------------------------------------------------------------------
    ViewConf(name="", attr="default", context=IRoot),
    ViewConf(name="base", attr="base", context=IRoot, renderer=t + "base.pt"),
    ViewConf(name="open", attr="open", context=IRoot),
    ViewConf(name="search", attr="viewsearch", context=IRoot, renderer=t + "search.pt"),
    ViewConf(name="settings", attr="settings", context=IRoot, renderer=t + "settings.pt"),
    ViewConf(name="slot", attr="slot", context=IRoot, renderer=t + "settings.pt"),

    # ViewConf(name = "options",attr = "options", context = IRoot,    renderer = t+"options.pt"),
    ViewConf(name="tool", attr="tool", context=IRoot, renderer=t + "tool.pt"),

    # Object views -----------------------------------------------------------------------------------
    # view
    ViewConf(name="", attr="default", context=IObject),
    # container object view
    # ViewConf(name = "",     attr = "view",  context = IContainer, renderer = t+"list.pt"),
    ViewConf(name="viewlist", attr="viewlist", context=IContainer, renderer=t + "list.pt"),
    ViewConf(name="open", attr="open", context=IContainer),
    # non container object view
    # ViewConf(name = "",     attr = "view",  context = INonContainer, renderer = t+"view.pt"),
    ViewConf(name="view", attr="view", context=INonContainer, renderer=t + "view.pt"),
    ViewConf(name="open", attr="open", context=INonContainer),

    # object pages
    ViewConf(name="edit", attr="edit", context=IObject, renderer=t + "edit.pt", permission="edit"),
    ViewConf(name="meta", attr="meta", context=IObject, renderer=t + "meta.pt"),
    ViewConf(name="options", attr="options", context=IObject, renderer=t + "options.pt"),
    ViewConf(name="file", attr="file", context=IObject),
    ViewConf(name="deleteo", attr="delete", context=IObject, renderer=t + "delete.pt", permission="delete"),
    ViewConf(name="@delfile", attr="delfile", context=IObject, permission="delete"),

    ViewConf(name="wf", attr="wf", context=IObject, renderer=t + "wf.pt", permission="edit"),
    ViewConf(name="wfa", attr="action", context=IObject, renderer=t + "form-workflow.pt", permission="edit"),
    ViewConf(name="wfa-i", attr="action", context=IObject, renderer=t + "form-workflow-inline.pt", permission="edit"),
    ViewConf(name="wft", attr="transition", context=IObject, permission="edit"),

    # container add, delete
    ViewConf(name="add", attr="add", context=IContainer, renderer=t + "add.pt", permission="add"),
    ViewConf(name="deletec", attr="deletelist", context=IContainer, renderer=t + "delete.pt", permission="delete"),
    ViewConf(name="duplicate", attr="duplicate", context=IObject, permission="add"),
    ViewConf(name="tool", attr="tool", context=IObject, renderer=t + "tool.pt", permission="tools"),

    # cut, copy, paste
    ViewConf(name="@cut",   attr = "cut",   context = IContainer, permission="edit"),
    ViewConf(name="@copy",  attr = "copy",  context = IContainer, permission="edit"),
    ViewConf(name="@paste", attr = "paste", context = IContainer, permission="add"),

    ViewConf(name="profile", attr="profile", context=IRoot, renderer=t + "profile.pt", permission="iface"),
]


class IFaceView(Parts, Search, CopyView, BaseView):
    """
    Basic peak interface view implementation
    Provides default functionality for data management and interface configuration.
    
    search and parts are required for html rendering. For Configuration an instance of IFace
    is loaded from context.
    
    Views:
    - default
    - open
    - Object view
    - Object edit form
    - Object duplicate
    - Object meta field list
    - Object workflow
    - Object options
    - Object tool
    - Container contents list
    - Container subitem delete
    - Container subitem add
    - Root base
    - Root settings
    
    Permissions:
    - iface_view
    - iface_edit
    - iface_delete
    - iface_add
    - iface_wf
    - iface_tools 
    """
    
    actionid = None
    
    def __init__(self, context, request):
        BaseView.__init__(self, context, request)
        self.ifaceConf = self.configuration.ifaceConf
        # default static routing for iface assets
        self.static = self.ifaceConf.get("static", "")
        if ":" in self.static:
            self.static = self.StaticUrl(self.static)

    # Get conf values -------------------------------------------------------------------

    def GetFlds(self, flds, container, addtype):
        flds = self._ResolveFlds(flds, object=None, addtype=addtype, addHidden=1, addReadonly=0)
        return self._LoadFldsConf(flds, container, addtype)


    def GetFldsAdd(self, container, addtype):
        flds = []
        typedef = self.context.app.configurationQuery.GetObjectConf(addtype)
        if addtype in self.ifaceConf.addflds:
            flds = self.ifaceConf.addflds[addtype]["fields"]
        else:
            conf = self._GetFirstByInterfaces(typedef, self.ifaceConf.addflds)
            if conf:
                flds = conf["fields"]
        if not flds:
            # use type configuration
            if typedef.get("forms") and typedef.forms.get("create"):
                flds = typedef.forms["create"].get("fields", [])
        flds = self._ResolveFlds(flds, object=None, addtype=addtype, addHidden=1, addReadonly=0)
        return self._LoadFldsConf(flds, container, addtype)
    
    
    def GetFldsEdit(self, object):
        flds = []
        if object.GetTypeID() in self.ifaceConf.editflds:
            flds = self.ifaceConf.editflds[object.GetTypeID()]["fields"]
        else:
            conf = self._GetFirstByInterfaces(object, self.ifaceConf.editflds)
            if conf:
                flds = conf["fields"]
        if not flds:
            # use type configuration
            conf = object.configuration
            if conf.get("forms") and conf.forms.get("edit"):
                flds = conf.forms["edit"].get("fields", [])
        flds = self._ResolveFlds(flds, object=object, addtype=None, addHidden=1, addReadonly=0)
        return self._LoadFldsConf(flds, object)
    
    
    def GetFldsMetaView(self, object):
        flds = []
        if not self.ifaceConf.metaflds:
            return flds
        if object.GetTypeID() in self.ifaceConf.metaflds:
            flds = self.ifaceConf.metaflds[object.GetTypeID()]["fields"]
        else:
            conf = self._GetFirstByInterfaces(object, self.ifaceConf.metaflds)
            if not conf:
                return flds
            flds = conf["fields"]
        flds = self._ResolveFlds(flds, object=object, addtype=None, addHidden=0, addReadonly=1)
        return self._LoadFldsConf(flds, object)


    def GetFldsDataView(self, object, files=None):
        flds = []
        if not self.ifaceConf.dataflds:
            return flds
        if object.GetTypeID() in self.ifaceConf.dataflds:
            flds = self.ifaceConf.dataflds[object.GetTypeID()]["fields"]
        else:
            conf = self._GetFirstByInterfaces(object, self.ifaceConf.dataflds)
            if not conf:
                return flds
            flds = conf["fields"]
        flds = self._ResolveFlds(flds, object=object, addtype=None, addHidden=0, addReadonly=1)
        flds = self._LoadFldsConf(flds, object)
        if files!=None:
            flds2 = []
            if files:
                for f in flds:
                    if f.datatype=="file":
                        flds2.append(f)
            else:
                for f in flds:
                    if f.datatype != "file":
                        flds2.append(f)
            flds=flds2
        return flds

    
    def GetTabs(self, object):
        tabs = self._GetAllByInterfaces(object, self.ifaceConf.tabs)
        return tabs
    
    
    def GetSettings(self, object):
        tabs = self._GetAllByInterfaces(object, self.ifaceConf.settings)
        return tabs


    def GetShortcuts(self, object):
        sc = self._GetAllByInterfaces(object, self.ifaceConf.shortcuts)
        return sc
    

    def GetSearchConf(self, searchid, container):
        if searchid and str(searchid) in self.ifaceConf.searchconf:
            c = self.ifaceConf.searchconf[searchid]
        else:
            c = self._GetFirstByInterfaces(container, self.ifaceConf.searchconf)
        if c:
            c["fields"] = self._GetFldsList(c["fields"], container, c["type"])
            c["form"] = self._GetFldsSearch(c["form"], container, c["type"])
            return c
        return {}


    # Redirects and navigation ------------------------------------------------------------

    def GetRedirect(self, action, view):
        if action == "default":
            v = "view"
            for d in self.ifaceConf.defaultviews:
                if self._ProvidesOne(view.context, d.get("interfaces")):
                    v = d["id"]
                    break
            return view.FolderUrl(view.context)+v
        
        elif action == "duplicate":
            return view.FolderUrl(view.context)+"edit"

        elif action == "deletelist":
            ref = self.GetFormValue("ref")
            return ref or view.PageUrl(view.context)

        return "parent"
    

    # current url / section -------------------------------------------------------------------------
    
    def IsInCurrentTree(self, item):
        """
        Check if navigation item is in current hierarchy or current page
        
        returns 0: outside, 1: in hierarchy, 2: current page 
        """
        url = self.request.url
        id = self.ToUrl(item)
        if url == id:
            return 2
        elif url.startswith(id):
            return 1
        return 0

    
    # field handling ---------------------------------------------------
    
    def _ProvidesOne(self, object, ifaces):
        if not ifaces:
            return False
        for i in ifaces:
            try:
                if isinstance(i, str):
                    i = ResolveName(i)
            except:
                continue
            if i.providedBy(object):
                return True
        return False

        
    def _GetFirstByInterfaces(self, object, conflist):
        """
        object may be object instance or object type id
        """
        if isinstance(object, str):
            object = self.context.app.configurationQuery.GetObjectConf(object)
        if IObjectConf.providedBy(object):
            object = self.context.root.factory.VirtualObj(object)
        if hasattr(conflist, "keys"):
            for k in list(conflist.keys()):
                if not IConf.providedBy(conflist[k]):
                    continue
                ifs = conflist[k].get("interfaces")
                if self._ProvidesOne(object, ifs):
                    return conflist[k]
        else:
            for k in conflist:
                if not IConf.providedBy(k):
                    continue
                ifs = k.get("interfaces")
                if self._ProvidesOne(object, ifs):
                    return k
        return None
    

    def _GetAllByInterfaces(self, object, conflist):
        l=[]
        if hasattr(conflist, "keys"):
            for k in list(conflist.keys()):
                if not IConf.providedBy(conflist[k]):
                    continue
                ifs = conflist[k].get("interfaces")
                if not ifs:
                    l.append(conflist[k])
                    continue
                if self._ProvidesOne(object, ifs):
                    l.append(conflist[k])
        else:
            for k in conflist:
                if not IConf.providedBy(k):
                    continue
                ifs = k.get("interfaces")
                if not ifs:
                    l.append(k)
                    continue
                if self._ProvidesOne(object, ifs):
                    l.append(k)
        return l

    
    def _ResolveFlds(self, flist, object=None, addtype=None, addHidden=1, addReadonly=0):
        """
        Resolve field field list and expand 'AllTypeDataFlds' and 'AllTypeMetaFlds' with 
        real fields.
        """
        cq = self.context.app.configurationQuery
        flds = []
        for f in flist:
            if f == AllTypeDataFlds:
                if addtype:
                    typefields = cq.GetAllObjectFlds(addtype)
                else:
                    typefields = cq.GetAllObjectFlds(object.GetTypeID())
                for f2 in typefields:
                    if not addReadonly and f2.get("readonly",False):
                        continue
                    if not addHidden and f2.get("hidden",False):
                        continue
                    flds.append(f2)

            elif f == AllTypeMetaFlds:
                for f2 in cq.GetAllMetaFlds():
                    if not addReadonly and f2.get("readonly",False):
                        continue
                    if not addHidden and f2.get("hidden",False):
                        continue
                    flds.append(f2)
            else:
                flds.append(f)
        return flds

    
    def _LoadFldsConf(self, fields, object, pool_type=None):
        """
        Lookup field configurations for all fields listed with field.id
        """
        if not pool_type:
            try:
                pool_type = object.GetTypeID()
            except:
                pass
        confs = []
        cq = self.context.app.configurationQuery
        for f in fields:
            if not isinstance(f, str):
                confs.append(f)
                continue
            f = cq.GetFld(f, pool_type)
            if f:
                confs.append(f)
        return confs


    def _GetFldsList(self, listflds, container, listtype):
        cq = self.context.app.configurationQuery
        flds = []
        for f in listflds:
            if f == AllTypeDataFlds:
                for f2 in cq.GetAllObjectFlds(listtype):
                    if not f2.get("readonly",False):
                        flds.append(f2)
            else:
                flds.append(f)
        return self._LoadFldsConf(flds, container, listtype)
    
    def _GetFldsSearch(self, searchflds, container, listtype):
        cq = self.context.app.configurationQuery
        flds = []
        for f in searchflds:
            if f == AllTypeDataFlds:
                for f2 in cq.GetAllObjectFlds(listtype):
                    if not f2.get("readonly",False):
                        flds.append(f2)
            else:
                flds.append(f)
        return self._LoadFldsConf(flds, container, listtype)
    

    def _AddObject(self, typeID, title, context=None, url="objid_url"):
        context = context or self.context
        if not context.IsTypeAllowed(typeID, self.User()):
            raise ValueError("Not allowed")

        form = ObjectForm(view=self, context=context, loadFromType=context.app.configurationQuery.GetObjectConf(typeID), autofill="off")
        form.subsets = {"create": {"fields": self.GetFldsAdd(context, typeID), "actions": ["default", "create", "cancel"]}}
        form.use_ajax = False
        form.Setup(subset="create", addTypeField=True)
        result, data, action = form.Process(redirectSuccess=url, redirectCancel=self.FolderUrl(context))
        if result and not data:
            self.Redirect(self.FolderUrl(context))
            return
        return dict(content=data, result=result, view=self, head=form.HTMLHead(), title=title)


    def _AddObjectIframe(self, typeID, title, context=None, url="objid_url"):
        context = context or self.context
        if not context.IsTypeAllowed(typeID, self.User()):
            raise ValueError("Not allowed")

        form = ObjectForm(view=self, context=context, loadFromType=context.app.configurationQuery.GetObjectConf(typeID), autofill="off")
        form.subsets = {"create": {"fields": self.GetFldsAdd(context, typeID), "actions": ["default", "create"]}}
        form.use_ajax = False
        form.Setup(subset="create", addTypeField=True)
        result, data, action = form.Process(renderSuccess=False)
        if result and action.id=="create":
            info = """
            <input type="hidden" class="addformresult" name="id" value="%d">
            <input type="hidden" class="addformresult" name="title" value="%s">
            <br>
            <p class="alert">Sie können den neuen Eintrag "%s" jetzt mit dem Button "Übernehmen" ins Formular übertragen.</p>
            """ % (result.id, result.meta.title, result.meta.title)
            return dict(content=data+info, result=result, view=self, head="", title=title)
        return dict(content=data, result=result, view=self, head=form.HTMLHead(), title=title)


    # helper -----------------------------------------------------------
    
    def ResolveUrl(self, url, context=None):
        """
        Resolve a string to url for object or the current context.

        Possible values:
        
        - page_url
        - obj_url
        - obj_folder_url
        - parent_url
        """
        if not url:
            return ""
        if context is None:
            context = self.context
        if url == "view_url":
            return self.FolderUrl(context)
        elif url == "section_url":
            try:
                section = self.request.currentSection
            except:
                section = ""
            return self.FolderUrl(context) + section
        elif url == "page_url":
            url = self.PageUrl(context)
        elif url == "obj_url":
            url = self.Url(context)
        elif url.find("objid_url")!=-1:
            url = url.replace("objid_url", self.FolderUrl(context.parent) + str(context.id))
        elif url.find("obj_folder_url")!=-1:
            url = url.replace("obj_folder_url", self.FolderUrl(context))
        elif url == "root_url":
            url = self.FolderUrl(context.root)
        elif url == "parent_url":
            url = self.Url(context.parent)
        return self.ToUrl(url)
    
        
    def ToUrl(self, item, context=None):
        """
        convert iface configuration item id to url
        replaced values: context., root., parent., pool., portal.
        """
        if not context:
            context = self.context
        if isinstance(item, str):
            u = item
        else:
            u = item["id"]
        if u.find("context.")!=-1:
            u = u.replace("context.", self.FolderUrl(context))
        elif u.find("root.")!=-1:
            u = u.replace("root.", self.FolderUrl(context.root))
        elif u.find("parent.")!=-1:
            u = u.replace("parent.", self.FolderUrl(context.parent))
        elif u.find("app.")!=-1:
            u = u.replace("app.", self.FolderUrl(context.app))
        elif u.find("portal.")!=-1:
            u = u.replace("portal.", self.FolderUrl(context.app.portal))
        return u

    def GetAction(self):
        """
        get the current action id from view or request
        """
        return self.actionid or self.GetFormValue("a")
    
    
    def IFaceCacheHeader(self, response):
        """
        Add the iface pages default cache header `configuration.cache`
        """
        if self.ifaceConf.cache=="no-cache":
            self.NoCache(response)
        self.Modified(response)
        

    def Form(self, conf, callback, formkws, tag):
        typeID = conf.get("loadFromType")
        form = HTMLForm(view=self, loadFromType=self.context.app.configurationQuery.GetObjectConf(typeID), autofill="off")
        form.fields = self.GetFlds(conf.fields, self.context, typeID)
        form.actions = [
            Conf(id="default",    method="StartForm",             name="Initialize",   hidden=True),
            Conf(id=tag,          method="ReturnDataOnSuccess",   name="Submit",       hidden=False),
        ]
        form.use_ajax = False
        form.ListenEvent("success", callback)
        form.Setup()
        result, data, action = form.Process(**formkws)
        if result and not data:
            self.Redirect(self.PageUrl(self.context))
            return
        return {"content": data, "result": result, "view": self, "foot": form.HTMLHead(), "pageinfo": tag}

    
    # default view callables ------------------------------------------------------------

    def slot(self):
        return {}

    def settings(self):
        return {}

    def tmpl(self):
        return {}

    def base(self):
        self.request.currentSection = "base"
        return {}


    # merged as one class for easier subclassing
    # class ObjectIFace(IFaceView, Parts):
    
    def default(self):
        url = self.GetRedirect("default", self)
        self.Redirect(url)

    def open(self):
        id = self.GetFormValue("id")
        v = self.GetFormValue("v")
        obj = None
        if id:
            if not IContainer.providedBy(self.context):
                obj = self.context.parent.GetObj(id)
            else:
                obj = self.context.GetObj(id)
            if obj is None:
                root = self.context.root
                obj = root.LookupObj(id)
        if not obj:
            return HTTPNotFound()
        url = self.FolderUrl(obj)
        self.Redirect(url)

    def meta(self):
        self.request.currentTab = "context.meta"
        return {}

    def view(self):
        self.request.currentTab = "context.view"
        return {"pageinfo": "view", "msgs": []}

    def viewlist(self):
        self.request.currentTab = "context.viewlist"
        searchid = self.context.configuration.id
        if not searchid or not str(searchid) in self.ifaceConf.searchconf:
            if IRoot.providedBy(self.context):
                searchid = "root"
                if not str(searchid) in self.ifaceConf.searchconf:
                    searchid = None
            else:
                searchid = None
        return {"pageinfo": "view", "msgs": [], "searchconf": searchid or "default"}

    def viewsearch(self):
        self.request.currentTab = "context.search"
        self.request.currentSection = "search"
        return dict(pageinfo="search", msgs=[], searchconf="search")

    def options(self):
        self.request.currentTab = "context.options"
        app = self.context.app
        selected = self.GetFormValue("tag")
        if selected:
            tool = app.GetTool(selected, contextObject=self.context)
            data = self.RenderView(tool)
            # pyramid bug? reset the active view in request
            self.request.__dict__['__view__'] = self
            return {"content": data, "tools": [], "tool": tool, "pageinfo": "options"}

        t = app.configurationQuery.GetAllToolConfs(contextObject=self.context)
        return {"content": "", "tools": t, "tool": None, "pageinfo": "options"}

    def add(self):
        self.request.currentTab = "context.add"
        typeID = self.GetFormValue("pool_type")
        if not typeID:
            return {"content": self.selectType(), "result": True}
            #raise ValueError, "No type id found"
        if not self.context.IsTypeAllowed(typeID, self.User()):
            raise ValueError("Not allowed")

        form = ObjectForm(view=self, loadFromType = self.context.app.configurationQuery.GetObjectConf(typeID), autofill="off")
        form.subsets = {"create": {"fields": self.GetFldsAdd(self.context, typeID), "actions": ["default","create"]}}
        form.use_ajax = False
        form.Setup(subset="create", addTypeField=True)
        result, data, action = form.Process(redirectSuccess="view_url")
        if result and not data:
            self.Redirect(self.PageUrl(self.context))
            return
        return {"content": data, "result": result, "view": self, "foot": form.HTMLHead(), "pageinfo": "add"}

    def edit(self):
        self.request.currentTab = "context.edit"
        form = ObjectForm(loadFromType = self.context.configuration, view=self, autofill="off")
        form.subsets = {"edit": {"fields": self.GetFldsEdit(self.context), "actions": ["edit"], "defaultAction": "defaultEdit"}}
        form.Setup(subset="edit")
        result, data, action = form.Process(redirectSuccess="view_url")
        return {"content": data, "result": result, "foot": form.HTMLHead(), "pageinfo": "edit"}

    def form(self):
        self.request.currentTab = "context.options"
        typeID = self.GetFormValue("pool_type")
        if not typeID:
            return {"content": self.selectType(), "result": True}
            #raise ValueError, "No type id found"
        if not self.context.IsTypeAllowed(typeID, self.User()):
            raise ValueError("Not allowed")

        form = ObjectForm(view=self, loadFromType = self.context.app.configurationQuery.GetObjectConf(typeID), autofill="off")
        form.subsets = {"create": {"fields": self.GetFldsAdd(self.context, typeID), "actions": ["default","create"]}}
        form.use_ajax = False
        form.Setup(subset="create", addTypeField=True)
        result, data, action = form.Process(redirectSuccess="view_url")
        if result and not data:
            self.Redirect(self.PageUrl(self.context))
            return
        return {"content": data, "result": result, "view": self, "foot": form.HTMLHead(), "pageinfo": "add"}

    def duplicate(self):
        url = self.GetRedirect("duplicate", self)
        self.Redirect(url)

    def deletelist(self):
        self.request.currentTab = "context.options"
        ids = self.GetFormValue("ids")
        if isinstance(ids, str):
            if "," in ids:
                ids = ids.split(",")
            else:
                ids = [ids]
        result = {"msgs": [], "objsToDelete": [], "action": "deletec", "result": False, "ids": ids, "pageinfo": "delete"}
        if not ids:
            result["msgs"] = [translator()(_("Nothing to delete"))]
            return result
        delete = self.GetFormValue("delete")
        objs = []
        for i in ids:
            o = self.context.obj(i)
            if o is not None:
                # check delete permission
                if self.Allowed("delete", o):
                    objs.append(o)
                else:
                    result["msgs"].append(o.meta.title + " " + translator()(_("Unauthorized. Delete is not allowed.")))
        if delete != "1":
            result["objsToDelete"] = objs
            return result
        ok = False
        user = self.User()
        for o in objs:
            # linked as unit/unitlist
            linked = self.context.root.search.GetReferences(o.id, types=None, sort="id")
            if linked:
                result["msgs"].append(o.meta.title + ": " + translator()(_("Still linked. Not deleted.")) + " " + str([l[0] for l in linked]))
                ok = False
                continue
            try:
                ok = self.context.Delete(o.id, user=user, obj=o)
            except Exception as e:
                result["msgs"].append(str(e))
                ok = False
        if not ok:
            return result
        url = self.GetRedirect("deletelist", self)
        self.Redirect(url, messages=[_("OK")])


    def delete(self, url="view_url"):
        self.request.currentTab = "context.options"
        delete = self.GetFormValue("delete")
        result = {"msgs": [], "objsToDelete": [self.context], "action": "deleteo", "result": False, "ids":None, "pageinfo": "delete"}
        if delete != "1":
            return result
        ok = False
        parent = self.context.parent
        try:
            ok = parent.Delete(self.context.id, user=self.User(), obj=self.context)
        except Exception as e:
            result["msgs"].append(str(e))
            ok = False
        if not ok:
            return result
        self.Redirect(self.ResolveUrl(url, parent), messages=[translator()(_("OK. Deleted."))])

    def delfile(self):
        self.ResetFlashMessages()
        file = self.GetFormValue("fid")
        user = self.User()
        try:
            r = self.context.DeleteFile(file, user)
            if not r:
                m = translator()(_("Something went wrong!"))
            else:
                m =  translator()(_("OK. Deleted."))
        except Exception as e:
            m=str(e)
        r = Response(content_type="text/html", conditional_response=True)
        r.text = "<span>%(msg)s</span>" % {"name": str(file), "msg":m}
        return r


    def action(self):
        action = self.GetFormValue("action")
        url = self.GetFormValue("ref")
        if not url:
            url = self.FolderUrl()
        values = dict()
        context = self.context
        user = self.User()

        # interactive transitions mit form auslesen und verarbeiten. nur 1 form pro transition.
        workflow = context.workflow
        process = workflow.GetWf()
        transitions = process.PossibleTransitions(context.meta.pool_wfa, action=action, context=context, user=user)
        if len(transitions)>=1:
            t = transitions[0]
            fncs = t.GetInteractiveFunctions()
            if fncs:
                fnc = fncs[0]
                if hasattr(fnc, "__init__") and fnc.form is None:
                    form = fnc(self).form
                else:
                    form = fnc
                form.Setup()
                form.widget.item_template = "field_onecolumn"
                form.widget.action_template = "form_actions_onecolumn"
                result, data, formaction = form.Process(renderSuccess=False, redirectCancel=url)
                if not result or (result and formaction.id != "submit"):
                    return dict(result=result, head=form.HTMLHead(), content=data, transition=t)
                values = result

        try:
            # workflow ausführen. form daten in 'data' wenn verarbeitet.
            transition = workflow.WfAction(action, user, values=values)
        except WorkflowNotAllowed:
            self.Redirect(url, messages=["Bereits ausgeführt"])
            return
        self.context.CommitInternal(user)
        msg = transition.message or "OK"
        self.Redirect(url, messages=[msg])

    def transition(self):
        transition = self.GetFormValue("t")
        url = self.GetFormValue("ref")
        if not url:
            url = self.Url()
        user = self.User()
        self.context.workflow.WfAction("", user, transition=transition)
        self.context.CommitInternal(user)
        msg = "OK"
        self.Redirect(url, messages=[msg])


    def wf(self):
        self.request.currentTab = "context.options"
        return {"pageinfo": "wf"}

    def tool(self):
        self.request.currentTab = "context.options"
        return {"pageinfo": "tool"}
    
    def profile(self):
        from nive_userdb.userview.view import UserForm
        from nive.security import IAdminUser

        user = self.User(sessionuser=False)
        if IAdminUser.providedBy(user):
            return dict(content="Singned in as System Admin", result=False, title="Profile")

        userdb = self.context.app.portal.userdb
        typeconf=userdb.configurationQuery.GetObjectConf("user")
        form = UserForm(view=self, context=user, loadFromType=typeconf)
        form.Setup(subset="edit")
        result, data, action = form.Process(renderSuccess=True)
        return dict(content=data, result=result, title="Profile")

    def selectType(self):
        user = self.User()
        lt = self.context.GetAllowedTypes(user)
        tmpl = """<p><a href="add?pool_type=%s" class="btn btn-primary add">%s</a></p>"""
        html = ""
        for t in lt:
            html += tmpl % (t["id"], t["name"])
        return html

    




# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = """
LocalGroups extension module
----------------------------
Security extension to handle local group assginments for users.
Can be used for Roots and Objects or any other python class supporting events
and id attribute (number). Uses idhash for root objects.
"""

from nive.definitions import ModuleConf, Conf
from nive.definitions import implementer, ILocalGroups


@implementer(ILocalGroups)
class LocalGroups(object):
    """
    """
    _owner = "group:owner"
    _secid = None

    def Init(self):
        self._localRoles = {}
        self.ListenEvent("create", "AddOwner")
        self.ListenEvent("delete", "RemoveGroups")
        self._secid = self._secid or self.id or self.idhash

    @property
    def securityID(self):
        return self._secid
    
    def GetLocalGroups(self, username, user=None):
        """
        Group assignments use the user name.
        returns a list of all local user groups, including parent settings
        """
        if self.id <= 0:
            return self._LocalGroups(username)
        g = []
        o = self
        while o:
            g += o._LocalGroups(username)
            o = o.parent
        return g 


    def AllLocalGroups(self):
        """
        returns a list of all local user group settings as list including 
        [username, group, id]. This function does not include parent level
        settings.
        """
        return self.db.GetGroups(self._secid)


    def AddOwner(self, user, **kw):
        """
        Add the current user as group:owner to local roles
        """
        if not user or not str(user):
            return
        self.AddLocalGroup(str(user), self._owner)
    
        
    def AddLocalGroup(self, username, group):
        """
        Add a local group assignment for username.
        """
        groups = self._LocalGroups(username)
        if group in groups:
            return 
        if username is None:
            return
        self._AddLocalGroupsCache(username, group)
        self.db.AddGroup(self._secid, userid=username, group=group)

        
    def RemoveLocalGroups(self, username, group=None):
        """
        Remove a local group assignment. If group is None all local groups
        will be removed.
        """
        self._DelLocalGroupsCache(username, group)
        self.db.RemoveGroups(self._secid, userid=username, group=group)


    def RemoveGroups(self, **kw):
        """
        Remove all group assignments before deleting the object. 
        """
        self.db.RemoveGroups(self._secid)
        self._localRoles = {}
        

    def _LocalGroups(self, username):
        if username in self._localRoles:
            return list(self._localRoles[username])
        g = [r[1] for r in self.db.GetGroups(self._secid, userid=username)]
        self._localRoles[username] = tuple(g)
        return g
    
    def _AddLocalGroupsCache(self, username, group):
        if username in self._localRoles:
            if group in self._localRoles[username]:
                return
            l = list(self._localRoles[username])
            l.append(group)
            self._localRoles[username] = tuple(l)
            return 
        self._localRoles[username] = (group,)
    
    def _DelLocalGroupsCache(self, username, group=None):
        if not username in self._localRoles:
            return
        if username in self._localRoles and not group:
            del self._localRoles[username]
            return
        if not group in self._localRoles[username]:
            return
        l = list(self._localRoles[username])
        l.remove(group)
        self._localRoles[username] = tuple(l)



def SetupLocalGroups(app, pyramidConfig):
    # get all roots and add extension
    extension = "nive.extensions.localgroups.LocalGroups"
    def add(confs):
        for c in confs:
            e = c.extensions
            if e and extension in e:
                continue
            if e is None:
                e = []
            if isinstance(e, tuple):
                e = list(e)
            e.append(extension)
            c.unlock()
            c.extensions = tuple(e)
            c.lock()
    
    add(app.configurationQuery.GetAllRootConfs())
    add(app.configurationQuery.GetAllObjectConfs())
    
    
configuration = ModuleConf(
    id = "localGroups",
    name = "Local Group assignment for objects and roots",
    context = "nive.extensions.localgroups",
    events = (Conf(event="startRegistration", callback=SetupLocalGroups),),
)



# -*- coding: latin-1 -*-

import unittest

from time import time

from nive.utils.dataPool2.tests.test_Base import stdMeta, struct, data1_1, data2_1, meta1, file1_1, file1_2




class dbTest(object):

    def statdb(self):
        c = self.pool.GetCountEntries()
        #print "Count entries in DB:", c
        return c

    # entries ---------------------------------------------------------------------------
    def create1(self):
        #print "Create Entry",
        e=self.pool.CreateEntry("data1", user="unittest")
        e.Commit(user="unittest")
        #print e.GetID(), "OK"
        return e.GetID()

    def create2(self):
        #print "Create Entry",
        e=self.pool.CreateEntry("data2", user="unittest")
        e.Commit(user="unittest")
        #print e.GetID(), "OK"
        return e.GetID()

    def get(self, id):
        #print "Load entry", id
        e=self.pool.GetEntry(id)
        #print "OK"
        return e

    # set ---------------------------------------
    def set1(self, id):
        #print "Store data", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e)
        e.data.update(data1_1)
        e.meta.update(meta1)
        e.Commit(user="unittest")
        self.assertTrue(e.GetMeta())
        d=e.GetData()
        self.assertTrue(d.get("ftext")    == data1_1.get("ftext")    )
        self.assertTrue(d.get("fnumber")    == data1_1.get("fnumber")    )
        self.assertTrue(self.pool.GetDBDate(str(d.get("fdate"))) == self.pool.GetDBDate(str(data1_1.get("fdate"))))
        self.assertTrue(d.get("flist")    == data1_1.get("flist")    )
        self.assertTrue(d.get("fmselect") == data1_1.get("fmselect") )
        self.assertTrue(d.get("funit")    == data1_1.get("funit")    )
        self.assertTrue(d.get("funitlist")== data1_1.get("funitlist"))
        self.assertTrue(d.get("ftext"))
        #print "OK"

    def set2(self, id):
        #print "Store data", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e)
        e.data.update(data2_1)
        e.meta.update(meta1)
        e.Commit(user="unittest")
        self.assertTrue(e.GetMeta())
        d=e.GetData()
        self.assertTrue(d.get("ftext")    == data2_1.get("ftext")    )
        self.assertTrue(d.get("fstr")     == data2_1.get("fstr")    )
        self.assertTrue(d.get("ftext"))
        #print "OK"

    def setfile1(self, id):
        #print "Store file", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e)
        self.assertTrue(e.CommitFile("file1", {"file":file1_1, "filename":"file1.txt"}))
        e.Commit(user="unittest")
        self.assertTrue(e.GetFile("file1").read() == file1_1)
        #print "OK"

    def setfile2(self, id):
        #print "Store file", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e)
        self.assertTrue(e.CommitFile("file2", {"file":file1_2, "filename":"file2.txt"}))
        e.Commit(user="unittest")
        self.assertTrue(e.GetFile("file2").read() == file1_2)
        #print "OK"


    # get --------------------------------------------------------------------------
    def data1(self, id):
        #print "Check entry data", id,
        e=self.pool.GetEntry(id)
        d=e.GetData()
        self.assertTrue(d.get("ftext")    == data1_1.get("ftext")    )
        self.assertTrue(d.get("fnumber")    == data1_1.get("fnumber")    )
        self.assertTrue(self.pool.GetDBDate(str(d.get("fdate"))) == self.pool.GetDBDate(str(data1_1.get("fdate"))))
        self.assertTrue(d.get("flist")    == data1_1.get("flist")    )
        self.assertTrue(d.get("fmselect") == data1_1.get("fmselect") )
        self.assertTrue(d.get("funit")    == data1_1.get("funit")    )
        self.assertTrue(d.get("funitlist")== data1_1.get("funitlist"))
        self.assertTrue(d.get("ftext"))
        #print "OK"

    def data2(self, id):
        #print "Check entry data", id,
        e=self.pool.GetEntry(id)
        d=e.GetData()
        self.assertTrue(d.get("ftext")    == data2_1.get("ftext")    )
        self.assertTrue(d.get("fstr")     == data2_1.get("fstr")    )
        self.assertTrue(d.get("ftext"))
        #print "OK"

    def file1(self, id):
        #print "Load file", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e.GetFile("file1").read() == file1_1)
        #print "OK"

    def file2(self, id):
        #print "Load file", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e.GetFile("file2").read() == file1_2)
        #print "OK"

    def fileErr(self, id):
        #print "Load non existing file", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e.GetFile("file1") == None)
        #print "OK"

    # getstream --------------------------------------------------------------------------
    def file1stream(self, id):
        #print "Load file", id,
        e=self.pool.GetEntry(id)
        s=e.GetFile("file1")
        d = s.read()
        s.close()
        self.assertTrue(d == file1_1)
        #print "OK"

    def file2stream(self, id):
        #print "Load file", id,
        e=self.pool.GetEntry(id)
        s=e.GetFile("file2")
        d=s.read()
        s.close()
        self.assertTrue(d == file1_2)
        #print "OK"

    # functions ------------------------------------------------------------------------
    def stat(self, id):
        e=self.pool.GetEntry(id)
        self.assertTrue(e.GetMetaField("pool_createdby")=="unittest")
        self.assertTrue(e.GetMetaField("pool_changedby")=="unittest")
        #print "Create: %s by %s    Changed: %s by %s" % (e.GetMetaField("pool_create"), e.GetMetaField("pool_createdby"), e.GetMetaField("pool_change"), e.GetMetaField("pool_changedby"))

    def delete(self, id):
        #print "Delete", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e)
        del e
        self.assertTrue(self.pool.DeleteEntry(id))
        self.pool.Commit(user="unittest")

    def duplicate(self, id,file=True):
        #print "Duplicate", id,
        e=self.pool.GetEntry(id)
        n=e.Duplicate(file)
        self.assertTrue(n)
        n.Commit(user="unittest")
        #print "OK"
        return n.GetID()

    def filetest(self, id):
        #print "File test", id,
        e=self.pool.GetEntry(id)
        self.assertTrue(e)

        self.assertSetEqual(set(e.FileKeys()), {"file1","file2"})

        self.assertTrue(e.GetFile("file1").filename=="file1.txt")
        self.assertTrue(e.GetFile("file2").filename=="file2.txt")
        self.assertTrue(e.GetFile("file1").filename)
        self.assertTrue(e.GetFile("file2").filename)

        self.assertTrue(e.GetFile("file1"))
        self.assertTrue(e.GetFile("file2"))
        self.assertTrue(e.GetFile("file3")==None)
        self.assertTrue(e.GetFile("")==None)

        l=e.Files({})
        self.assertTrue(len(l)==2)
        l2=[]
        for f in l:
            l2.append(f["filekey"])
        self.assertTrue("file1" in l2)
        self.assertTrue("file2" in l2)
        #print "OK"





    # new test fncs -----------------------------------------------------------------------------

    def test_create_empty(self):

        t = time()
        c = self.statdb()
        # creating
        id1=self.create1()
        e=self.get(id1)
        del e
        self.stat(id1)

        id2=self.create2()
        e=self.get(id2)
        del e
        self.stat(id2)

        # cnt    ok
        c2 = self.statdb()
        self.assertTrue(c+2==c2)

        # deleting
        self.delete(id1)
        self.delete(id2)
        c3 = self.statdb()
        self.assertTrue(c==c3)




    def test_create_base(self):

        t = time()
        c = self.statdb()
        # creating
        id1=self.create1()
        e=self.get(id1)
        del e
        self.stat(id1)

        id2=self.create2()
        e=self.get(id2)
        del e
        self.stat(id2)

        # cnt    ok
        c2 = self.statdb()
        self.assertTrue(c+2==c2)

        #update data
        self.set1(id1)
        self.set2(id2)
        self.setfile1(id1)
        self.setfile2(id1)

        #load data
        self.data1(id1)
        self.data2(id2)
        self.file1(id1)
        self.file2(id1)

        # deleting
        self.delete(id1)
        self.delete(id2)
        c3 = self.statdb()
        self.assertTrue(c==c3)



    def test_files_base(self):

        t = time()
        c = self.statdb()

        # creating
        id1=self.create1()
        e=self.get(id1)

        #update data
        self.set1(id1)
        self.setfile1(id1)
        self.setfile2(id1)

        #load data
        self.file1(id1)
        self.file2(id1)

        # file
        self.filetest(id1)

        # deleting
        self.delete(id1)
        c3 = self.statdb()
        self.assertTrue(c==c3)


    def test_preload(self):

        t = time()
        self.statdb()

        id=self.create1()

        #print "Preload Skip", id,
        e = self.pool.GetEntry(id, preload="skip")
        self.assertTrue(e.GetDataRef()>0 and e.GetDataTbl()!="")
        del e
        #print "OK"

        #print "Preload Meta", id,
        e = self.pool.GetEntry(id, preload="meta")
        self.assertTrue(e.GetDataRef()>0 and e.GetDataTbl()!="")
        del e
        #print "OK"

        #print "Preload All", id,
        e = self.pool.GetEntry(id, preload="all")
        self.assertTrue(e.GetDataRef()>0 and e.GetDataTbl()!="")
        del e
        #print "OK"

        #print "Preload MetaData", id,
        e = self.pool.GetEntry(id, preload="metadata")
        self.assertTrue(e.GetDataRef()>0 and e.GetDataTbl()!="")
        del e
        #print "OK"

        #print "Preload StdMeta", id,
        e = self.pool.GetEntry(id, preload="stdmeta")
        self.assertTrue(e.GetDataRef()>0 and e.GetDataTbl()!="")
        del e
        #print "OK"

        #print "Preload StdMetaData", id,
        e = self.pool.GetEntry(id, preload="stdmetadata")
        self.assertTrue(e.GetDataRef()>0 and e.GetDataTbl()!="")
        del e
        #print "OK"

        self.delete(id)
        self.assertTrue(self.pool.IsIDUsed(id) == False)


    def test_duplicate_base(self):

        t = time()
        c=self.statdb()

        # creating
        id1=self.create1()
        id2=self.create2()

        # cnt    ok
        c2 = self.statdb()
        self.assertTrue(c+2==c2)

        #update data
        self.set1(id1)
        self.set2(id2)
        self.setfile1(id1)
        self.setfile2(id1)

        #load data
        self.data1(id1)
        self.data2(id2)
        self.file1(id1)
        self.file2(id1)

        # duplicate
        id3=self.duplicate( id1)
        id4=self.duplicate( id2)
        id5=self.duplicate( id1, file=False)

        #load dupl. data
        self.data1(id3)
        self.data1(id5)
        self.data2(id4)
        self.file1(id3)
        self.file2(id3)
        self.fileErr(id5)

        # deleting
        self.delete(id1)
        self.delete(id2)
        self.delete(id3)
        self.delete(id4)
        self.delete(id5)
        c3 = self.statdb()
        self.assertTrue(c==c3)


    def test_sql(self):

        t = time()
        sql, values=self.pool.FmtSQLSelect(list(stdMeta)+list(struct["data1"]),
                        {"pool_type": "data1", "ftext": "123", "fnumber": 300000},
                        sort = "title, id, fnumber",
                        ascending = 0,
                        dataTable = "data1",
                        operators={"pool_type":"=", "ftext": "<>", "fnumber": "<"},
                        start=1,
                        max=123)
        self.pool.Query(sql, values)
        c=self.pool.Execute(sql, values)
        c.close()
        #print "OK"

        sql, values=self.pool.FmtSQLSelect(list(struct["data1"]),
                                     {"ftext": "", "fnumber": 3},
                                     dataTable="data1",
                                     sort = "id, fnumber",
                                     ascending = 1,
                                     operators={"ftext": "=", "fnumber": "="},
                                     start=1,
                                     max=123,
                                     singleTable=1)
        self.pool.Query(sql, values)
        c=self.pool.Execute(sql, values)
        c.close()
        #print "OK"

        #print "GetFulltextSQL",
        sql, values=self.pool.GetFulltextSQL("is",
                            list(stdMeta)+list(struct["data1"]),
                            {},
                            sort = "title",
                            ascending = 1,
                            dataTable = "data1")
        self.pool.Query(sql, values)
        c=self.pool.Execute(sql, values)
        c.close()
        #print "OK"


    def test_sql2(self):

        t = time()
        sql1, values1=self.pool.FmtSQLSelect(list(stdMeta)+list(struct["data1"]),
                        {"pool_type": "data1", "ftext": "", "fnumber": 3},
                        sort = "title, id, fnumber",
                        ascending = 0,
                        dataTable = "data1",
                        operators={"pool_type":"=", "ftext": "<>", "fnumber": ">"},
                        start=1,
                        max=123)
        sql2, values2=self.pool.FmtSQLSelect(list(struct["data1"]),
                                     {"ftext": "", "fnumber": 3},
                                     dataTable="data1",
                                     sort = "id, fnumber",
                                     ascending = 1,
                                     operators={"ftext": "<>", "fnumber": ">"},
                                     start=1,
                                     max=123,
                                     singleTable=1)
        sql3, values3=self.pool.GetFulltextSQL("is",
                            list(stdMeta)+list(struct["data1"]),
                            {},
                            sort = "title",
                            ascending = 1,
                            dataTable = "data1")
        c=self.pool.connection.cursor()
        c.execute(sql1, values1)
        c.execute(sql2, values2)
        c.execute(sql3, values3)
        
        self.pool.SelectFields("data1", fields=("id",), idValues=[0], idColumn="id")
        self.pool.SelectFields("pool_meta", fields=("id","title","pool_type"), idValues=[1,2,3,4,5], idColumn="pool_unitref")


    def test_insertdelete(self):
        self.pool.DeleteRecords("pool_meta", {"pool_type": "notype", "title": "test entry"})
        self.pool.Commit()

        self.pool.InsertFields("pool_meta", {"pool_type": "notype", "title": "test entry"})
        self.pool.Commit()
        sql, values = self.pool.FmtSQLSelect(["id"], {"pool_type": "notype", "title": "test entry"}, dataTable="pool_meta", singleTable=1) 
        id = self.pool.Query(sql, values)
        self.assertTrue(id)
        
        self.pool.UpdateFields("pool_meta", id[0][0], {"pool_type": "notype 123", "title": "test entry 123"})
        self.pool.Commit()
        sql, values = self.pool.FmtSQLSelect(["id"], {"pool_type": "notype", "title": "test entry"}, dataTable="pool_meta", singleTable=1) 
        id = self.pool.Query(sql, values)
        self.assertFalse(id)
        sql, values = self.pool.FmtSQLSelect(["id"], {"pool_type": "notype 123", "title": "test entry 123"}, dataTable="pool_meta", singleTable=1) 
        id = self.pool.Query(sql, values)
        self.assertTrue(id)
        
        for i in id:
            self.pool.DeleteRecords("pool_meta", {"id":i[0]})
        self.pool.Commit()
        sql, values = self.pool.FmtSQLSelect(["id"], {"pool_type": "notype 123", "title": "test entry 123"}, dataTable="pool_meta", singleTable=1) 
        id = self.pool.Query(sql, values)
        self.assertFalse(id)


    def test_groups(self):
        userid = "123"
        group = "group:test"
        id = 1
        ref = "o"
        self.pool.RemoveGroups(id=id)
        self.assertFalse(self.pool.GetGroups(id, userid, group))
        self.assertFalse(self.pool.GetGroups(id))
        self.assertFalse(self.pool.GetGroups((id,id)))
        self.pool.AddGroup(id, userid, group)
        self.assertTrue(self.pool.GetGroups(id, userid, group))
        self.assertTrue(self.pool.GetGroups(id))
        self.assertTrue(self.pool.GetGroups((id,id)))

        self.pool.RemoveGroups(userid=userid, group=group, id=id)
        self.assertFalse(self.pool.GetGroups(id, userid, group))


    def test_search_files(self):

        t = time()
        c = self.statdb()

        # creating
        id1=self.create1()
        id2=self.create1()
        id3=self.create1()

        #update data
        self.set1(id1)
        self.setfile1(id1)
        self.setfile2(id1)

        self.set1(id2)
        self.setfile1(id2)
        self.setfile2(id2)

        self.set1(id3)
        self.setfile1(id3)
        self.setfile2(id3)

        dbfile = self.pool
        #print "SearchFilename",
        f1 = dbfile.SearchFilename("file1.txt")
        #print len(f1),
        self.assertTrue(len(f1)>=3)
        f2 = dbfile.SearchFilename("file2.txt")
        #print len(f2),
        self.assertTrue(len(f2)>=3)
        f3 = dbfile.SearchFilename("fileXXX.txt")
        #print len(f3),
        self.assertTrue(len(f3)==0)

        #print "SearchFiles",
        parameter={"id": (id1,id2,id3)}
        operators={"id": "IN"}
        f = dbfile.SearchFiles(parameter, operators=operators) #sort="filename",
        #print len(f),
        self.assertTrue(len(f)==6)
        parameter["filename"] = "file2.txt"
        operators["filename"] = "="
        f = dbfile.SearchFiles(parameter, operators=operators) #sort="size",
        #print len(f),
        self.assertTrue(len(f)==3)
        #print "OK"

        # deleting
        self.delete(id1)
        self.delete(id2)
        self.delete(id3)
        c3 = self.statdb()
        self.assertTrue(c==c3)


    def test_tree(self):
        base = self.pool
        #base.GetContainedIDs(base=0, sort="title", parameter="")
        #base.GetTree(flds=["id"], sort="title", base=0, parameter="")
        base.GetParentPath(1)
        base.GetParentTitles(1)





